//rww - blood stuff

#include "quakedef.h"

extern int hr_blood1;
extern int hr_blood2;
extern int hr_blood3;
extern int hr_blood4;

extern cvar_t cl_blobtime;

void GL_RotationMatrix(float phi, int axis, float *mat);
pmtrace_t PM_GeneralTrace(vec3_t start, vec3_t end, qboolean thirdHull);

#define BLOOD_BLOB_GRAV		4.0f
#define MAX_BLOOD_BLOBS		512
#define NUM_BLOB_VERTS		12
#define NUM_BLOB_TRIS		14
typedef struct bloodBlob_s
{
	float		rotMat[16];
	float		center[3];
	float		centerVel[3];
	qboolean	centerHitFloor;
	float		verts[NUM_BLOB_VERTS*3];
	float		vertVel[NUM_BLOB_VERTS*3];
	float		vertSpeed[NUM_BLOB_VERTS];
	double		createTime;
	double		lifeSpan;
	double		fadeTime;
	qboolean	hitFloor[NUM_BLOB_VERTS];
	int			tex;
	float		trailPoint[3];
	float		trailDir[3];
	float		unique1;
	float		unique2;
	float		unique3;
	qboolean	alive;
} bloodBlob_t;

static qboolean g_bloodInit = false;

static bloodBlob_t g_bloodBlobs[MAX_BLOOD_BLOBS];

static int g_blobTris[NUM_BLOB_TRIS*3];
static float g_blobTC[NUM_BLOB_VERTS*2];
static float g_blobOffsets[NUM_BLOB_VERTS*3];
static const float g_blobSpread = 3.0f;//6.0f;
static const float g_blobHeightBase = 0.0f;
static const float g_blobHeightElevated = 3;//6.0f;

//init blood stuff
void Blood_Init(void)
{
	float *off;
	int *tri;
	int i;
	
	g_bloodInit = true;

	off = g_blobOffsets;
	//0
	off[0] = -g_blobSpread;
	off[1] = 0.0f;
	off[2] = g_blobHeightBase;
	off += 3;
	//1
	off[0] = -(g_blobSpread*0.65f);
	off[1] = -(g_blobSpread*0.5f);
	off[2] = g_blobHeightBase;
	off += 3;
	//2
	off[0] = 0.0f;
	off[1] = -g_blobSpread;
	off[2] = g_blobHeightBase;
	off += 3;
	//3
	off[0] = (g_blobSpread*0.65f);
	off[1] = -(g_blobSpread*0.5f);
	off[2] = g_blobHeightBase;
	off += 3;
	//4
	off[0] = g_blobSpread;
	off[1] = 0.0f;
	off[2] = g_blobHeightBase;
	off += 3;
	//5
	off[0] = (g_blobSpread*0.65f);
	off[1] = (g_blobSpread*0.5f);
	off[2] = g_blobHeightBase;
	off += 3;
	//6
	off[0] = 0.0f;
	off[1] = g_blobSpread;
	off[2] = g_blobHeightBase;
	off += 3;
	//7
	off[0] = -(g_blobSpread*0.65f);
	off[1] = (g_blobSpread*0.5f);
	off[2] = g_blobHeightBase;
	off += 3;

	//8
	off[0] = -(g_blobSpread*0.25f);
	off[1] = -(g_blobSpread*0.25f);
	off[2] = g_blobHeightElevated;
	off += 3;
	//9
	off[0] = (g_blobSpread*0.25f);
	off[1] = -(g_blobSpread*0.25f);
	off[2] = g_blobHeightElevated;
	off += 3;
	//10
	off[0] = (g_blobSpread*0.25f);
	off[1] = (g_blobSpread*0.25f);
	off[2] = g_blobHeightElevated;
	off += 3;
	//11
	off[0] = -(g_blobSpread*0.25f);
	off[1] = (g_blobSpread*0.25f);
	off[2] = g_blobHeightElevated;
	off += 3;

	tri = g_blobTris;
	//0
	tri[0] = 0;
	tri[1] = 1;
	tri[2] = 8;
	tri += 3;
	//1
	tri[0] = 1;
	tri[1] = 2;
	tri[2] = 8;
	tri += 3;
	//2
	tri[0] = 8;
	tri[1] = 2;
	tri[2] = 9;
	tri += 3;
	//3
	tri[0] = 2;
	tri[1] = 3;
	tri[2] = 9;
	tri += 3;
	//4
	tri[0] = 9;
	tri[1] = 3;
	tri[2] = 4;
	tri += 3;
	//5
	tri[0] = 10;
	tri[1] = 9;
	tri[2] = 4;
	tri += 3;
	//6
	tri[0] = 10;
	tri[1] = 4;
	tri[2] = 5;
	tri += 3;
	//7
	tri[0] = 6;
	tri[1] = 10;
	tri[2] = 5;
	tri += 3;
	//8
	tri[0] = 11;
	tri[1] = 10;
	tri[2] = 6;
	tri += 3;
	//9
	tri[0] = 7;
	tri[1] = 11;
	tri[2] = 6;
	tri += 3;
	//10
	tri[0] = 0;
	tri[1] = 11;
	tri[2] = 7;
	tri += 3;
	//11
	tri[0] = 0;
	tri[1] = 8;
	tri[2] = 11;
	tri += 3;

	//12
	tri[0] = 8;
	tri[1] = 9;
	tri[2] = 11;
	tri += 3;
	//13
	tri[0] = 11;
	tri[1] = 9;
	tri[2] = 10;
	tri += 3;

	for (i = 0; i < NUM_BLOB_VERTS; i++)
	{
		float *tc = &g_blobTC[i*2];
		float *off = &g_blobOffsets[i*3];
		tc[0] = ((off[0]+g_blobSpread)*0.5f)/g_blobSpread;
		tc[1] = ((off[1]+g_blobSpread)*0.5f)/g_blobSpread;
	}
}

//interpoalte from a set of values
float Blood_GetMappedValue(float perc, float *values, int numValues)
{
	const float frac = 1.0f/(numValues-1);
	float stage;
	int iStage;
	float stageAmt;

	if (perc < 0.0f)
	{
		perc = 0.0f;
	}
	else if (perc > 1.0f)
	{
		perc = 1.0f;
	}
	stage = perc/frac;
	iStage = (int)stage;
	stageAmt = (perc*(numValues-1))-iStage;

	return (values[iStage]*(1.0f-stageAmt))+(values[iStage+1]*stageAmt);
}

//render a single blob
void Blood_RenderBlob(bloodBlob_t *b)
{
	int i;
	float fade = 1.0f;
	float trailLife = 1.5f;
	if (b->lifeSpan < trailLife)
	{
		trailLife = b->lifeSpan;
	}

	if (b->lifeSpan != 0 && b->fadeTime != 0)
	{ //check for a time-based fadeout
		double lifeDif = b->lifeSpan-b->fadeTime;
		double lifePoint = realtime-b->createTime;
		if (lifePoint > lifeDif)
		{
			lifePoint -= lifeDif;
			fade = 1.0f-(lifePoint/b->fadeTime);
		}
	}
	if (fade < 0.0f)
	{
		fade = 0.0f;
	}

	GL_Bind(b->tex);
	glDisable(GL_CULL_FACE);

	glEnable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBlendFunc(GL_DST_COLOR, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f(fade, fade, fade, fade);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), b->verts);
	glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GL_FLOAT), g_blobTC);
	glDrawElements(GL_TRIANGLES, NUM_BLOB_TRIS*3, GL_UNSIGNED_INT, g_blobTris);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

	//render the trail
	if (b->createTime+(trailLife) >= realtime)
	{
		float w = 8.0f;
		float h = 8.0f;
		float orthoMat[16];
		float t[3];
		double corner1 = realtime*2;
		double corner2 = realtime*3;
		double corner3 = realtime*4;
		double corner4 = realtime*5;
		float cornerFactor = 1.0f;
		float trav;
		float endDist = 22 + (b->unique1*8.0f);
		float endSize = 16.0f + (b->unique2*8.0f);
		float interp[8] = {0.0f, 0.3f, 0.5f, 0.7f, 0.8f, 0.9f, 0.95f, 1.0f};

		fade = 1.0f;
		if (b->fadeTime != 0)
		{ //check for a time-based fadeout
			double lifeDif = 0.0f;//(b->lifeSpan*trailFrac)-b->fadeTime;
			double lifePoint = realtime-b->createTime;
			if (lifePoint > lifeDif)
			{
				lifePoint -= lifeDif;
				fade = 1.0f-(lifePoint/(trailLife));
			}
		}
		if (fade < 0.0f)
		{
			fade = 0.0f;
		}

		w += (1.0f-fade)*endSize;
		h += (1.0f-fade)*endSize;

		//glBlendFunc(GL_ONE, GL_ONE);

		trav = Blood_GetMappedValue(1.0f-fade, interp, 8);
		t[0] = b->trailPoint[0] + b->trailDir[0]*(trav*endDist);
		t[1] = b->trailPoint[1] + b->trailDir[1]*(trav*endDist);
		t[2] = b->trailPoint[2] + b->trailDir[2]*(trav*endDist);
		VectorSubtract(t, r_refdef.vieworg, t);
		glPushMatrix();
		glTranslatef(r_refdef.vieworg[0], r_refdef.vieworg[1], r_refdef.vieworg[2]);
		glTranslatef(t[0], t[1], t[2]);
		glRotatef (-r_refdef.viewangles[1],  0, 0, -1);
		glRotatef (-r_refdef.viewangles[0],  0, -1, 0);
		glRotatef (-r_refdef.viewangles[2],  -1, 0, 0);

		glRotatef ((b->unique3 * 1024.0f) - (fade*80.0f),  1, 0, 0);
		glRotatef (-90.0f,  0, 1, 0);
		glRotatef (-90.0f,  0, 0, 1);

		GL_Bind(hr_blood4);
		glColor4f(fade, 0.0f, 0.0f, fade);
		glBegin(GL_QUADS);
			glTexCoord2f(0.0f, 1.0f);
			glVertex2f(-(w*0.5f)+(sinf(corner1)*cornerFactor), -(h*0.5f)+(cosf(corner1)*cornerFactor));
			glTexCoord2f(0.0f, 0.0f);
			glVertex2f(-(w*0.5f)+(sinf(corner2)*cornerFactor), (h*0.5f)+(cosf(corner2)*cornerFactor));
			glTexCoord2f(1.0f, 0.0f);
			glVertex2f((w*0.5f)+(sinf(corner3)*cornerFactor), (h*0.5f)+(cosf(corner3)*cornerFactor));
			glTexCoord2f(1.0f, 1.0f);
			glVertex2f((w*0.5f)+(sinf(corner4)*cornerFactor), -(h*0.5f)+(cosf(corner4)*cornerFactor));
		glEnd();

		glPopMatrix();
	}

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable(GL_BLEND);

	glEnable(GL_CULL_FACE);
}

//simulate a single point
//fixme, this function became ugly
void Blood_SimulatePoint(bloodBlob_t *b, int i, float *p, float *v, float *vertSpeed, float *center, float timeFrac)
{
	float oldPos[3];
	pmtrace_t trace;

	if (center)
	{
		trace = PM_GeneralTrace(center, p, false);
		if (!trace.startsolid && trace.fraction != 1.0f)
		{
			VectorCopy(trace.endpos, p);
			p[0] += trace.plane.normal[0];
			p[1] += trace.plane.normal[1];
			p[2] += trace.plane.normal[2];
		}
	}

	VectorCopy(p, oldPos);
	p[0] += v[0]*((*vertSpeed)*timeFrac);
	p[1] += v[1]*((*vertSpeed)*timeFrac);
	p[2] += v[2]*((*vertSpeed)*timeFrac);
	v[0] *= (1.0f-(timeFrac*0.1f));
	v[1] *= (1.0f-(timeFrac*0.1f));
	v[2] *= (1.0f-(timeFrac*0.1f));
	if ((i == -1 && !b->centerHitFloor) || !b->hitFloor[i])
	{
		v[2] -= BLOOD_BLOB_GRAV*timeFrac;
	}
	trace = PM_GeneralTrace(oldPos, p, false);
	if (trace.startsolid)
	{ //remove
		b->createTime = 0;
		return;
	}
	if (trace.fraction != 1.0f)
	{
		float dp = DotProduct(v, trace.plane.normal);
		float up[3] = {0.0f, 0.0f, 1.0f};
		float u = DotProduct(up, trace.plane.normal);
		v[0] -= dp*trace.plane.normal[0];
		v[1] -= dp*trace.plane.normal[1];
		v[2] -= dp*trace.plane.normal[2];
		if (u > -0.7f && fabsf(u) <= 0.6f)
		{
			v[2] = 0.0f;
			if (i >= 8)
			{
				(*vertSpeed) *= 0.25f;
			}
			else
			{
				(*vertSpeed) *= 0.5f;
			}
		}
		/*
		if (center && u > -0.7f)
		{ //let blood slide down certain planes slower
			float test[3];
			pmtrace_t testtr;
			test[0] = center[0]-trace.plane.normal[0]*16.0f;
			test[1] = center[1]-trace.plane.normal[1]*16.0f;
			test[2] = center[2]-trace.plane.normal[2]*16.0f;
			testtr = PM_GeneralTrace(center, test, false);
			if (!testtr.startsolid && testtr.fraction != 1.0f)
			{ //near the wall on the direction of the plane
				float hug = 1024.0f;
				(*vertSpeed) *= 0.75f;
				if (fabsf(u) <= 0.7f)
				{
					if (i >= 8)
					{
						(*vertSpeed) *= 0.15f;
					}
					else
					{
						(*vertSpeed) *= 0.75f;
					}
				}
				//let the rest of the blob cling to this surface
				b->centerVel[0] -= trace.plane.normal[0]*hug;
				b->centerVel[1] -= trace.plane.normal[1]*hug;
				b->centerVel[2] = -1024.0f;
			}
			else
			{
				(*vertSpeed) = 0.5f;
			}
		}
		*/
		if (u > 0.9f)
		{
			if (i == -1)
			{
				b->centerHitFloor = true;
			}
			else
			{
				b->hitFloor[i] = true;
			}
		}
		VectorCopy(trace.endpos, p);
		p[0] += trace.plane.normal[0];
		p[1] += trace.plane.normal[1];
		p[2] += trace.plane.normal[2];
	}
}

//blob physics
void Blood_SimulateBlob(bloodBlob_t *b, float timeFrac)
{
	int i;
	float f = 0.3f;
	Blood_SimulatePoint(b, -1, b->center, b->centerVel, &f, NULL, timeFrac);
	for (i = 0; i < NUM_BLOB_VERTS; i++)
	{
		float *p = &b->verts[i*3];
		float *v = &b->vertVel[i*3];
		float *off = &g_blobOffsets[i*3];
		float offRot[3];
		float rest[3];
		float d[3];
		Blood_SimulatePoint(b, i, p, v, &b->vertSpeed[i], b->center, timeFrac);

		TransformPointByMatrix4x4(b->rotMat, off, offRot);

		if (b->hitFloor[i])
		{
			offRot[2] *= 0.1f;
		}

		//constantly pull toward "rest" position
		rest[0] = b->center[0]+(offRot[0]*4.0f+(sinf(realtime*i)*0.5f));
		rest[1] = b->center[1]+(offRot[1]*4.0f+(cosf(realtime*i)*0.5f));
		rest[2] = b->center[2]+(offRot[2]*4.0f);
		VectorSubtract(rest, p, d);
		/*
		if (Length(d) > 32.0f)
		{
			VectorNormalize(d);
			d[0] *= 32.0f;
			d[1] *= 32.0f;
			d[2] *= 32.0f;
			VectorAdd(rest, d, p);
		}
		*/
		if (Length(d) > 4.0f)
		{
			v[0] += d[0]*0.05f + (timeFrac*0.1f);
			v[1] += d[1]*0.05f + (timeFrac*0.1f);
			v[2] += d[2]*0.05f + (timeFrac*0.1f);
		}
	}
}

//render active blobs
void Blood_RenderBlobs(void)
{
	int i;

	if (!g_bloodInit)
	{
		Blood_Init();
	}

	glDepthMask(GL_FALSE);
	for (i = 0; i < MAX_BLOOD_BLOBS; i++)
	{
		bloodBlob_t *b = &g_bloodBlobs[i];
		if (!b->alive)
		{
			continue;
		}
		if ((b->createTime+b->lifeSpan) >= realtime)
		{
			float f = host_frametime*30.0f;
			if (f > 2.0f)
			{
				f = 2.0f;
			}
			Blood_SimulateBlob(b, f);
			Blood_RenderBlob(b);
		}
		else
		{
			b->alive = false;
		}
	}
	glDepthMask(GL_TRUE);
}

//get the oldest blob
bloodBlob_t *Blood_OldestBlob(void)
{
	int i;
	bloodBlob_t *old = &g_bloodBlobs[0];
	for (i = 1; i < MAX_BLOOD_BLOBS; i++)
	{
		if (g_bloodBlobs[i].createTime < old->createTime)
		{
			old = &g_bloodBlobs[i];
		}
	}
	return old;
}

//add some blobs
void Blood_AddBlobs(float *origin, float *dir, int count, float cone)
{
	int i;
	int num;
	float ang[3];
	float trans[16];

	count *= 2;

	dir[2] += 0.6f;
	VectorNormalize(dir);

	vectoangles(dir, ang);

	glPushMatrix();
	glLoadIdentity();
	glRotatef(ang[2], 1, 0, 0);
	glRotatef(-ang[0], 0, 1, 0);
	glRotatef(ang[1], 0, 0, 1);

	glRotatef(90.0f, 0, 1, 0);

	glGetFloatv(GL_MODELVIEW_MATRIX, trans);
	glPopMatrix();

	if (!g_bloodInit)
	{
		Blood_Init();
	}

	for (num = 0; num < count; num++)
	{
		float r;
		float startVel = 32.0f;
		int rint;
		bloodBlob_t *b = Blood_OldestBlob();
		memset(b, 0, sizeof(bloodBlob_t));
		b->alive = true;

		b->unique1 = (float)(rand()%8192)/8192.0f;
		b->unique2 = (float)(rand()%8192)/8192.0f;
		b->unique3 = (float)(rand()%8192)/8192.0f;

		r = (float)(rand()%360);
		GL_RotationMatrix(r*0.0174532925f, 2, b->rotMat);

		VectorCopy(dir, b->centerVel);
		r = -(cone*0.5f) + ((rand()&0x7fff) / ((float)0x7fff))*cone;
		b->centerVel[0] += r;
		r = -(cone*0.5f) + ((rand()&0x7fff) / ((float)0x7fff))*cone;
		b->centerVel[1] += r;
		r = -(cone*0.5f) + ((rand()&0x7fff) / ((float)0x7fff))*cone;
		b->centerVel[2] += r;
		VectorNormalize(b->centerVel);

		VectorCopy(b->centerVel, b->trailDir);
		b->trailPoint[0] = origin[0] + b->trailDir[0]*16.0f;
		b->trailPoint[1] = origin[1] + b->trailDir[1]*16.0f;
		b->trailPoint[2] = origin[2] + b->trailDir[2]*16.0f;

		b->centerVel[0] *= startVel;
		b->centerVel[1] *= startVel;
		b->centerVel[2] *= startVel;

		rint = rand()%9;
		if (rint < 3)
		{
			b->tex = hr_blood1;
		}
		else if (rint < 6)
		{
			b->tex = hr_blood2;
		}
		else
		{
			b->tex = hr_blood3;
		}
		b->createTime = realtime;
		b->lifeSpan = cl_blobtime.value;
		b->fadeTime = cl_blobtime.value*0.8f;
		VectorCopy(origin, b->center);
		for (i = 0; i < NUM_BLOB_VERTS; i++)
		{
			float *v = &b->verts[i*3];
			float *vel = &b->vertVel[i*3];
			float t[3], sized[3];
			float *off = &g_blobOffsets[i*3];
			VectorCopy(off, sized);
			sized[0] *= 0.8f;
			sized[1] *= 0.8f;
			b->vertSpeed[i] = 0.3f;
			TransformPointByMatrix4x4(trans, sized, t);
			v[0] = origin[0]+(t[0]);
			v[1] = origin[1]+(t[1]);
			v[2] = origin[2]+(t[2]);
			t[0] += b->centerVel[0]*0.5f;
			t[1] += b->centerVel[1]*0.5f;
			t[2] += b->centerVel[2]*0.5f;
			VectorNormalize(t);
			vel[0] = t[0]*startVel;
			vel[1] = t[1]*startVel;
			vel[2] = t[2]*startVel;
		}
	}
}
