#define MAX_VOLUME_TRIS		20000
#define MAX_VOLUME_VERTS	20000
extern float volVerts[3*MAX_VOLUME_VERTS];
extern int volTris[3*MAX_VOLUME_TRIS];

#ifdef SERVERONLY
	#ifndef BYTE
	#define BYTE unsigned char
	#endif
	#ifndef WORD
	#define WORD unsigned short
	#endif
	#ifndef DWORD
	#define DWORD unsigned int
	#endif

typedef struct modelMatrix_s
{
	float				x1[3];
	float				x2[3];
	float				x3[3];
	float				o[3];
} modelMatrix_t;

typedef struct modelTan_s
{
	int					numNormals;
	modelMatrix_t		tangentMatrix;
} modelTan_t;

typedef struct modelTriNeighbor_s
{ //used occasionally for convenience
	int					a, b, c;
} modelTriNeighbor_t;

typedef struct modelVert_s
{ //used occasionally for convenience
	float				x, y, z;
} modelVert_t;

typedef struct modelTriFace_s
{ //used occasionally for convenience
	int					a, b, c;
} modelTriFace_t;

typedef struct modelTexCoord_s
{ //used occasionally for convenience
	float				u, v;
} modelTexCoord_t;
#endif

#pragma pack(1) //all of these structs are meant to be unpadded
typedef struct modelChunk_s
{
	WORD				type;
	DWORD				len;
	BYTE				data;
} modelChunk_t;

typedef struct modelVertWInfo_s
{
	int					weightIndex;
	int					numWeights;
} modelVertWInfo_t;

typedef struct modelRDMTriFace_s
{
	WORD				a;
	WORD				b;
	WORD				c;
	WORD				flag;
} modelRDMTriFace_t;

typedef struct modelTriFaceGLFormat_s
{
	int					a;
	int					b;
	int					c;
} modelTriFaceGLFormat_t;

typedef struct modelRGBA_s
{
	float				rgba[4];
} modelRGBA_t;


typedef struct modelRGB_s
{
	float				rgb[3];
} modelRGB_t;

typedef struct modelVertWeight_s
{
	int					index;
	int					boneIndex;
	float				weightFactor;
	float				pos[3];
} modelVertWeight_t;

typedef struct modelBone_s modelBone_s;
typedef struct modelBoneExData_s
{
	modelBone_s			*parent;

	modelMatrix_t		invBaseMat;
	modelMatrix_t		localMat;
	modelMatrix_t		parentMat;
	modelMatrix_t		finalMat;

	double				transformTime;
} modelBoneExData_t;

typedef struct modelBone_s
{
	int					index;
	char				name[32];
	
	modelMatrix_t		mat;
	char				parentName[32];

	modelBoneExData_t	eData;
} modelBone_t;

typedef struct triPlaneEq_s
{
	float p[4];
} triPlaneEq_t;

typedef struct boneOverrides_s
{
	modelMatrix_t		modifyMat;
} boneOverrides_t;

//data in here should be per-instance and not shared between models
typedef struct modelInstanceData_s
{
	modelVert_t					*vertDataTransformed;
	qboolean					vertDataDataCalculated;
	double						vertDataTransformedTime;
	qboolean					vertDataWasAnimating;
	float						vertDataLastScale[3];
	int							vertDataLastDeform;

	float						transObjMins[3];
	float						transObjMaxs[3];

	boneOverrides_t				*boneOverrides;
	qboolean					bonesModified;

	float						*shadVolVerts;
	int							*shadVolTris;
	int							numShadVolVerts;
	int							numShadVolTris;
	float						shadowVolMins[3];
	float						shadowVolMaxs[3];
	float						shadowVolScale;

#ifndef SERVERONLY
	modelDecal_t				*decals;
#endif

	unsigned int				glVertBufferIndex;
	unsigned int				glIndexBufferIndex;

	modelTan_t					*tangentData;
	qboolean					tangentDataCalculated;
	double						tangentDataCalculatedTime;
	qboolean					tangentWasAnimating;

	triPlaneEq_t				*clipTriPlaneEq;
	triPlaneEq_t				*triPlaneEq;
	qboolean					triPlDataCalculated;
	double						triPlDataCalculatedTime;
	qboolean					triPlWasAnimating;

	float						*vertNormalsTransformed;

	modelRGBA_t					*colorPointer;
	float						*lightVectors;
	float						*halfAngleVectors;
} modelInstanceData_t;

typedef struct modelTriMesh_s
{
	int							numVerts;
	BYTE						*vertData;
	int							numFaces;
	BYTE						*faceData;
	float						*vertNormals;
	BYTE						*vertNormalIndexes;
	char						skinName[1024];
	char						skinNameNormal[1024];

	int							numWeights;
	modelVertWeight_t			*vertWeights;
	int							numBones;
	modelBone_t					*bones;

	modelVertWInfo_t			*vertWInfo;

	BYTE						*faceMatList;
	BYTE						*uvCoord;
	BYTE						*smoothingGrp;
	BYTE						*coordOffset;

	//extra
	modelTriFaceGLFormat_t		*faceDataFormatted;
	modelTriNeighbor_t			*faceNeighbors;

	qboolean					vertNormalsStored;
	qboolean					inheritedBones;

	modelInstanceData_t			mid;
} modelTriMesh_t;

typedef struct modelHierData_s
{
	WORD				flag1;
	WORD				flag2;
	WORD				father;
} modelHierData_t;

typedef struct modelAnimTrackFrames_s
{
	DWORD				start;
	DWORD				end;
} modelAnimTrackFrames_t;

typedef struct modelAnimTrackKeyPos_s
{
	DWORD				keyNum;
	WORD				accelDataAvail;

	float				accelData; //this will trail off for keyNum.. or something?
	float				pos[3];
} modelAnimTrackKeyPos_t;

typedef struct modelAnimTrackKeyRot_s
{
	DWORD				keyNum;
	WORD				accelDataAvail;

	float				accelData; //this will trail off for keyNum.. or something?
	float				angle;
	float				axis[3];
} modelAnimTrackKeyRot_t;

typedef struct modelAnimTrackKeyScale_s
{
	DWORD				keyNum;
	WORD				accelDataAvail;

	float				accelData; //this will trail off for keyNum.. or something?
	float				scale[3];
} modelAnimTrackKeyScale_t;

typedef struct modelAnimTrack_Pos_s
{
	WORD						flag;

	DWORD						unknown1; //wts? 3ds spec says unknown for 8 bytes
	DWORD						unknown2;

	DWORD						keys;

	modelAnimTrackKeyPos_t		trackKeys;
} modelAnimTrack_Pos_t;

typedef struct modelAnimTrack_Rot_s
{
	WORD						flag;

	DWORD						unknown1; //wts? 3ds spec says unknown for 8 bytes
	DWORD						unknown2;

	DWORD						keys;

	modelAnimTrackKeyRot_t		trackKeys;
} modelAnimTrack_Rot_t;

typedef struct modelAnimTrack_Scale_s
{
	WORD						flag;

	DWORD						unknown1; //wts? 3ds spec says unknown for 8 bytes
	DWORD						unknown2;

	DWORD						keys;

	modelAnimTrackKeyScale_t	trackKeys;
} modelAnimTrack_Scale_t;

#define			OBJFLAG_OVERRIDE_MATRIX			(1<<1)
#define			OBJFLAG_DISABLE_SURFACE			(1<<2)

typedef struct modelMeshObject_s modelMeshObject_s;
typedef struct modelMeshObject_s
{
	char						*objName;
	modelTriMesh_t				meshData;
	modelHierData_t				*hierarchy;
	WORD						hierarchyPos;
	float						*pivotPoint;
	modelAnimTrack_Pos_t		*animTrack_Pos;
	modelAnimTrack_Rot_t		*animTrack_Rot;
	modelAnimTrack_Scale_t		*animTrack_Scale;

	modelMatrix_t				overrideMatrix;
	modelMatrix_t				curMatrix;
	int							flags;

	float						mins[3];
	float						maxs[3];

	int							gridCull;
	int							objTexture;
	int							objTextureNormal;

	qboolean					bloom;
	qboolean					alphaMask;
	qboolean					transparent;

	modelMeshObject_s			*parent;
	modelMeshObject_s			*next;
} modelMeshObject_t;


typedef struct rdaBone_s
{
	int							index;
	int							parentIndex;
	modelMatrix_t				*mat;
} rdaBone_t;

typedef struct rdaHeader_s
{
	int							version;
	int							numFrames;
	int							frameRate;

	int							numBones;
} rdaHeader_t;

typedef struct rda_s
{
	rdaHeader_t					hdr;
	rdaBone_t					*bones;
} rda_t;

typedef struct modelFile_s
{
	BYTE				*fileBuffer;
	int					fileLen;
	modelMeshObject_t	*firstObject;
	float				baseScale;
	rda_t				*baseAnim;
#ifndef SERVERONLY
	qboolean			lodHack;
	safeModelPtr_s		*lodModel;
	float				lodDist;
	qboolean			noShadows;
	qboolean			treeHack;
	float				treeBoneVel[3][3];
	float				treeBoneEff[3][3];
	double				treeSoundTime;
	int					treeSoundChan;
#endif
} modelFile_t;

#define MAX_PATH_CHARS 1024

typedef struct safeModelPtr_s
{
	modelFile_t		*p;
	char			fileName[MAX_PATH_CHARS];
	int				index;
	qboolean		allocated;
	qboolean		explicitlyFreed;
	int				userCount;
} safeModelPtr_t;

typedef struct drawObject_s drawObject_t;

typedef struct chunkTable_s
{
	int			chunkMain;
	int			chunkVersion;
	int			chunk3DEdit;
	int			chunkObject;
	int			chunkObjectTriMesh;
	int			chunkObjectVertData;
	int			chunkObjectFaceData;
	int			chunkObjectFaceMatData;
	int			chunkObjectTexCoords;
	int			chunkObjectSmoothGrp;
	int			chunkObjectCoordOffset;
	int			chunkKeyframer;
	int			chunkMeshInfo;
	int			chunkStartEndFrames;
	int			chunkHierarchy;
	int			chunkKeyframerHeader;
	int			chunkPivotPoint;
	int			chunkPositionTrack;
	int			chunkRotationTrack;
	int			chunkScaleTrack;
	int			chunkRollTrack;
	int			chunkHierarchyPos;

	int			chunkObjectVertNormals;
	int			chunkObjectVertWeights;
	int			chunkObjectVertWInfo;
	int			chunkObjectBoneNum;
	int			chunkObjectBone;
} chunkTable_t;

#define DEFORM_TEXWAVE						(1<<0)
#define DEFORM_TEXSCROLLX					(1<<1)
#define DEFORM_TEXSCROLLY					(1<<2)
#define DEFORM_MDLPULSE						(1<<3)
#define DEFORM_MDLPLANESCALE				(1<<4)
#define DEFORM_TOSCREEN						(1<<5)

typedef struct modelParmData_s
{
	rda_t				*anim;
	int					frame;
	int					oldFrame;
	float				frameLerp;

	float				modelScale[3];
	int					deforms;
	float				deformPropA;
	float				deformPropB;
	float				deformPropC;
	float				deformPropD;
} modelParmData_t;

#pragma pack(8) //back to default alignment

#ifndef SERVERONLY
void Model_DecalsInsert(modelMeshObject_t *obj, modelDecal_t *dec);
void Model_DecalsRemove(modelMeshObject_t *obj, modelDecal_t *dec);
#endif
void Model_FreeModel(modelMeshObject_t *rootObj, qboolean freeAsLL);
void Model_CacheInstance(const char *filename);
int Model_Precache(const char *filename);
safeModelPtr_t *Model_Allocate(const char *filename);
void Model_Free(int index);
void Model_ClearPool(qboolean setExplicit);
modelMeshObject_t *Model_GetObjectByName(const char *objName, modelFile_t *model);
void Model_CreateTransformedVerts(modelParmData_t *sceneObj, modelMeshObject_t *obj, qboolean doTrans, double colTrans);
void Model_TransformBones(modelMeshObject_t *obj, rda_t *anim, int frame, int oldFrame,
						 float lerpAmt, double timeCheck);
void Model_CreateFormattedFaces(modelMeshObject_t *obj);
void Model_ToggleRender(modelFile_t *model, const char *objName, qboolean visible);
void Model_ToggleRenderAllSurfs(modelFile_t *model, qboolean visible);
void Model_MatrixOverride(modelFile_t *model, const char *objName, modelMatrix_t *matrix);
void Model_GetModelInfo(safeModelPtr_t *model, int *numVerts, int *numFaces,
						float **vertData, int **triData);
void Model_GetModelInfoMultiObj(safeModelPtr_t *model, int *numVerts, int *numFaces,
						float **vertData, int **triData);
modelFile_t *SafeModelPtr_GetModel(safeModelPtr_t *safe);
void Model_LoadRDMForQuake(model_t *mod, unsigned char *buf);
modelBone_t *Model_FindObjectBoneByName(modelMeshObject_t *obj, const char *boneName);

#ifndef SERVERONLY
modelFile_t *Model_CoreFromQuakeEnt(entity_t *ent);
void Model_DrawRDMModel(entity_t *ent, int lightNum);
void GL_DrawRDMVolume(entity_t *ent, int lightNum);
#endif

rda_t *Anim_Allocate(const char *filename);
void Anim_ClearPool(void);
