/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"
#include "gl_modrdm.h" //rww

int			skytexturenum;

#ifndef GL_RGBA4
#define	GL_RGBA4	0
#endif

float g_beginLodCutoff = 2048.0f; //rww
float g_endLodCutoff = 4096.0f; //rww

extern qboolean g_depthPass; //rww

int		lightmap_bytes;		// 1, 2, or 4

int		lightmap_textures;

int		normalmap_default; //rww - flat normal texture
int		attenmap_default; //rww - default dlight 3d texture attenutation map

//rww
int		grass_textures[16];
int		grass_shadow_textures[16];
int		num_grass_textures = 0;

#define BLOCKLIGHT_NUM	512*512
unsigned		blocklights[BLOCKLIGHT_NUM]; //rww - boosted //[18*18];

//rww - boosted
#define	BLOCK_WIDTH		512 //128
#define	BLOCK_HEIGHT	512 //128

#define	MAX_LIGHTMAPS	128//64 //rww - bumped
int			active_lightmaps;

extern float g_sunPos[3]; //rww
extern qboolean g_hasSun; //rww

extern cvar_t gl_twoSidedStencil; //rww
extern cvar_t gl_specExp; //rww
extern cvar_t gl_showTris; //rww

qboolean GL_UploadPNG(char *filename, int tex, qboolean mips); //rww

qboolean GL_RDMModelInLight(entity_t *e, dlight_t *dlight); //rww
qboolean GL_AliasModelInLight(entity_t *e, dlight_t *dlight); //rww

void R_RandomPointInTri(float *v1, float *v2, float *v3, float *out); //rww

typedef struct glRect_s {
	unsigned char l,t,w,h;
} glRect_t;

glpoly_t	*lightmap_polys[MAX_LIGHTMAPS];
qboolean	lightmap_modified[MAX_LIGHTMAPS];
glRect_t	lightmap_rectchange[MAX_LIGHTMAPS];

int			allocated[MAX_LIGHTMAPS][BLOCK_WIDTH];

// the lightmap texture data needs to be kept in
// main memory so texsubimage can update properly
byte		lightmaps[4*MAX_LIGHTMAPS*BLOCK_WIDTH*BLOCK_HEIGHT];

// For gl_texsort 0
msurface_t  *skychain = NULL;
msurface_t  *waterchain = NULL;

void R_RenderDynamicLightmaps (msurface_t *fa);

void GL_CalcPolyLightVectors(glpoly_t *poly, dlight_t *dlight, entity_t *e, qboolean getHalfAngle);
void DrawGLPoly (texture_t *t, int lightmaptex, glpoly_t *p, dlight_t *dlight, entity_t *e, qboolean specPass);

#ifdef _DEBUG
//rdtsc for precise timing
void GL_TimingBegin(__int64 *time) {
	__asm {
		rdtsc					//stuff number of clock cycles since computer boot into eax+edx
		mov		ebx, time
		mov		[ebx], eax		//put eax in low 4 bytes
		mov		[ebx+4], edx	//put edx in high 4 bytes
	}
}

//rdtsc for precise timing
void GL_TimingEnd(__int64 *time) {
	__int64 prof_end = 0;
	__int64 *p = &prof_end;

	__asm {
		rdtsc					//stuff number of clock cycles since computer boot into eax+edx
		mov		ebx, p
		mov		[ebx], eax		//put eax in low 4 bytes
		mov		[ebx+4], edx	//put edx in high 4 bytes
	}

	*time = prof_end-(*time);
}
#endif

//rww - convenience
void GL_DlightHiLo(int lnum, int *low, int *high)
{
	*high = 1<<(lnum>>5);
	*low = 1<<(lnum%32);
}

//rww
static __inline qboolean GL_SurfLitByLight(msurface_t *surf, int lnum)
{
	if ( !(surf->dlightbits & (1<<(lnum%32)) ) )
		return false;

	if ( !(surf->dlightbitsHigh & (1<<(lnum>>5)) ) )
		return false;

	return true;
}

/*
===============
R_AddDynamicLights
===============
*/
void R_AddDynamicLights (msurface_t *surf)
{
	int			lnum;
	int			sd, td;
	float		dist, rad, minlight;
	vec3_t		impact, local;
	int			s, t;
	int			i;
	int			smax, tmax;
	mtexinfo_t	*tex;

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	tex = surf->texinfo;

	for (lnum=0 ; lnum<g_numLiveLights ; lnum++)
	{
		if ( !GL_SurfLitByLight(surf, g_liveLightsIdx[lnum]) )
			continue;		// not lit by this light

		rad = g_liveLights[lnum]->radius;
		dist = DotProduct (g_liveLights[lnum]->origin, surf->plane->normal) -
				surf->plane->dist;
		rad -= fabs(dist);
		minlight = g_liveLights[lnum]->minlight;
		if (rad < minlight)
			continue;
		minlight = rad - minlight;

		for (i=0 ; i<3 ; i++)
		{
			impact[i] = g_liveLights[lnum]->origin[i] -
					surf->plane->normal[i]*dist;
		}

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];
		
		for (t = 0 ; t<tmax ; t++)
		{
			td = local[1] - t*16;
			if (td < 0)
				td = -td;
			for (s=0 ; s<smax ; s++)
			{
				sd = local[0] - s*16;
				if (sd < 0)
					sd = -sd;
				if (sd > td)
					dist = sd + (td>>1);
				else
					dist = td + (sd>>1);
				if (dist < minlight)
					blocklights[t*smax + s] += (rad - dist)*256;
			}
		}
	}
}


/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the 8.8 format in blocklights
===============
*/
void R_BuildLightMap (msurface_t *surf, byte *dest, int stride)
{
	int			smax, tmax;
	int			t;
	int			i, j, size;
	byte		*lightmap;
	unsigned	scale;
	int			maps;
	unsigned	*bl;

	surf->cached_dlight = (surf->dlightframe == r_framecount);

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;
	size = smax*tmax;
	lightmap = surf->samples;

	if (size >= BLOCKLIGHT_NUM)
	{ //rww
		Sys_Error("R_BuildLightMap: size >= BLOCKLIGHT_NUM");
	}

// set to full bright if no light data
	if (/* r_fullbright.value || */ !cl.worldmodel->lightdata)
	{
		for (i=0 ; i<size ; i++)
			blocklights[i] = 255*256;
		goto store;
	}

// clear to no light
	for (i=0 ; i<size ; i++)
		blocklights[i] = 0;

// add all the lightmaps
	if (lightmap)
		for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ;
			 maps++)
		{
			scale = d_lightstylevalue[surf->styles[maps]];
			surf->cached_light[maps] = scale;	// 8.8 fraction
			for (i=0 ; i<size ; i++)
				blocklights[i] += lightmap[i] * scale;
			lightmap += size;	// skip to next lightmap
		}

// add all the dynamic lights
	if (!g_canUseTexFP || !gl_normalMaps.value)
	{ //rww - only if dlights are not realtime
		if (surf->dlightframe == r_framecount)
			R_AddDynamicLights (surf);
	}

// bound, invert, and shift
store:
	switch (gl_lightmap_format)
	{
	case GL_RGBA:
		stride -= (smax<<2);
		bl = blocklights;
		for (i=0 ; i<tmax ; i++, dest += stride)
		{
			for (j=0 ; j<smax ; j++)
			{
				t = *bl++;
				t >>= 7;
				if (t > 255)
					t = 255;
				dest[3] = 255-t;
				dest += 4;
			}
		}
		break;
	case GL_ALPHA:
	case GL_LUMINANCE:
	case GL_INTENSITY:
		bl = blocklights;
		for (i=0 ; i<tmax ; i++, dest += stride)
		{
			for (j=0 ; j<smax ; j++)
			{
				t = *bl++;
				t >>= 7;
				if (t > 255)
					t = 255;
				dest[j] = 255-t;
			}
		}
		break;
	default:
		Sys_Error ("Bad lightmap format");
	}
}


/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base)
{
	int		reletive;
	int		count;

	if (currententity->frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}
	
	if (!base->anim_total)
		return base;

	reletive = (int)(cl.time*10) % base->anim_total;

	count = 0;	
	while (base->anim_min > reletive || base->anim_max <= reletive)
	{
		base = base->anim_next;
		if (!base)
			Sys_Error ("R_TextureAnimation: broken cycle");
		if (++count > 100)
			Sys_Error ("R_TextureAnimation: infinite cycle");
	}

	return base;
}


/*
=============================================================

	BRUSH MODELS

=============================================================
*/


extern	int		solidskytexture;
extern	int		alphaskytexture;
extern	float	speedscale;		// for top sky and bottom sky

void DrawGLWaterPoly (glpoly_t *p);
void DrawGLWaterPolyLightmap (glpoly_t *p);

#ifdef _WIN32
lpMTexFUNC qglMTexCoord2fSGIS = NULL;
lpSelTexFUNC qglSelectTextureSGIS = NULL;
#endif

qboolean mtexenabled = false;

void GL_SelectTexture (GLenum target);

void GL_DisableMultitexture(void) 
{
	if (mtexenabled) {
		glDisable(GL_TEXTURE_2D);
		GL_SelectTexture(TEXTURE0_SGIS);
		mtexenabled = false;
	}
}

void GL_EnableMultitexture(void) 
{
	if (gl_mtexable) {
		GL_SelectTexture(TEXTURE1_SGIS);
		glEnable(GL_TEXTURE_2D);
		mtexenabled = true;
	}
}

#ifndef _WIN32
/*
================
R_DrawSequentialPoly

Systems that have fast state and texture changes can
just do everything as it passes with no need to sort
================
*/
void R_DrawSequentialPoly (msurface_t *s)
{
	glpoly_t	*p;
	float		*v;
	int			i;
	texture_t	*t;

	//
	// normal lightmaped poly
	//
//	if ((!(s->flags & (SURF_DRAWSKY|SURF_DRAWTURB)))
//		&& ((r_viewleaf->contents!=CONTENTS_EMPTY && (s->flags & SURF_UNDERWATER)) ||
//		(r_viewleaf->contents==CONTENTS_EMPTY && !(s->flags & SURF_UNDERWATER))))
	if (0)
	{
		p = s->polys;

		t = R_TextureAnimation (s->texinfo->texture);
		GL_Bind (t->gl_texturenum);
		glBegin (GL_POLYGON);
		v = p->verts[0];
		for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
		{
			glTexCoord2f (v[3], v[4]);
			glVertex3fv (v);
		}
		glEnd ();

		GL_Bind (lightmap_textures + s->lightmaptexturenum);
		glEnable (GL_BLEND);
		glBegin (GL_POLYGON);
		v = p->verts[0];
		for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
		{
			glTexCoord2f (v[5], v[6]);
			glVertex3fv (v);
		}
		glEnd ();

		glDisable (GL_BLEND);

		return;
	}

	//
	// subdivided water surface warp
	//
	if (s->flags & SURF_DRAWTURB)
	{
		GL_Bind (s->texinfo->texture->gl_texturenum);
		EmitWaterPolys (s);
		return;
	}

	//
	// subdivided sky warp
	//
	if (s->flags & SURF_DRAWSKY)
	{
		GL_Bind (solidskytexture);
		speedscale = realtime*8;
		speedscale -= (int)speedscale;

		EmitSkyPolys (s);

		glEnable (GL_BLEND);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		GL_Bind (alphaskytexture);
		speedscale = realtime*16;
		speedscale -= (int)speedscale;
		EmitSkyPolys (s);
		if (gl_lightmap_format == GL_LUMINANCE)
			glBlendFunc (GL_ZERO, GL_ONE_MINUS_SRC_COLOR);

		glDisable (GL_BLEND);
	}

	//
	// underwater warped with lightmap
	//
	p = s->polys;

	t = R_TextureAnimation (s->texinfo->texture);
	GL_Bind (t->gl_texturenum);
	DrawGLWaterPoly (p);

	GL_Bind (lightmap_textures + s->lightmaptexturenum);
	glEnable (GL_BLEND);
	DrawGLWaterPolyLightmap (p);
	glDisable (GL_BLEND);
}
#else
/*
================
R_DrawSequentialPoly

Systems that have fast state and texture changes can
just do everything as it passes with no need to sort
================
*/
void R_DrawSequentialPoly (msurface_t *s)
{
	glpoly_t	*p;
	float		*v;
	int			i;
	texture_t	*t;
	vec3_t		nv, dir;
	float		ss, ss2, length;
	float		s1, t1;
	glRect_t	*theRect;

	//
	// normal lightmaped poly
	//

	if (! (s->flags & (SURF_DRAWSKY|SURF_DRAWTURB|SURF_UNDERWATER) ) )
	{
		R_RenderDynamicLightmaps (s);
		if (gl_mtexable) {
			p = s->polys;

			t = R_TextureAnimation (s->texinfo->texture);
			// Binds world to texture env 0
			GL_SelectTexture(TEXTURE0_SGIS);
			GL_Bind (t->gl_texturenum);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
			// Binds lightmap to texenv 1
			GL_EnableMultitexture(); // Same as SelectTexture (TEXTURE1)
			GL_Bind (lightmap_textures + s->lightmaptexturenum);
			i = s->lightmaptexturenum;
			if (lightmap_modified[i])
			{
				lightmap_modified[i] = false;
				theRect = &lightmap_rectchange[i];
				glTexSubImage2D(GL_TEXTURE_2D, 0, 0, theRect->t, 
					BLOCK_WIDTH, theRect->h, gl_lightmap_format, GL_UNSIGNED_BYTE,
					lightmaps+(i* BLOCK_HEIGHT + theRect->t) *BLOCK_WIDTH*lightmap_bytes);
				theRect->l = BLOCK_WIDTH;
				theRect->t = BLOCK_HEIGHT;
				theRect->h = 0;
				theRect->w = 0;
			}
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
			glBegin(GL_POLYGON);
			v = p->verts[0];
			for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
			{
				qglMTexCoord2fSGIS (TEXTURE0_SGIS, v[3], v[4]);
				qglMTexCoord2fSGIS (TEXTURE1_SGIS, v[5], v[6]);
				glVertex3fv (v);
			}
			glEnd ();
			return;
		} else {
			p = s->polys;

			t = R_TextureAnimation (s->texinfo->texture);
			GL_Bind (t->gl_texturenum);
			glBegin (GL_POLYGON);
			v = p->verts[0];
			for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
			{
				glTexCoord2f (v[3], v[4]);
				glVertex3fv (v);
			}
			glEnd ();

			GL_Bind (lightmap_textures + s->lightmaptexturenum);
			glEnable (GL_BLEND);
			glBegin (GL_POLYGON);
			v = p->verts[0];
			for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
			{
				glTexCoord2f (v[5], v[6]);
				glVertex3fv (v);
			}
			glEnd ();

			glDisable (GL_BLEND);
		}

		return;
	}

	//
	// subdivided water surface warp
	//

	if (s->flags & SURF_DRAWTURB)
	{
		GL_DisableMultitexture();
		GL_Bind (s->texinfo->texture->gl_texturenum);
		EmitWaterPolys (s);
		return;
	}

	//
	// subdivided sky warp
	//
	if (s->flags & SURF_DRAWSKY)
	{
		GL_DisableMultitexture();
		GL_Bind (solidskytexture);
		speedscale = realtime*8;
		speedscale -= (int)speedscale & ~127;

		EmitSkyPolys (s);

		glEnable (GL_BLEND);
		GL_Bind (alphaskytexture);
		speedscale = realtime*16;
		speedscale -= (int)speedscale & ~127;
		EmitSkyPolys (s);

		glDisable (GL_BLEND);
		return;
	}

	//
	// underwater warped with lightmap
	//
	R_RenderDynamicLightmaps (s);
	if (gl_mtexable) {
		p = s->polys;

		t = R_TextureAnimation (s->texinfo->texture);
		GL_SelectTexture(TEXTURE0_SGIS);
		GL_Bind (t->gl_texturenum);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		GL_EnableMultitexture();
		GL_Bind (lightmap_textures + s->lightmaptexturenum);
		i = s->lightmaptexturenum;
		if (lightmap_modified[i])
		{
			lightmap_modified[i] = false;
			theRect = &lightmap_rectchange[i];
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, theRect->t, 
				BLOCK_WIDTH, theRect->h, gl_lightmap_format, GL_UNSIGNED_BYTE,
				lightmaps+(i* BLOCK_HEIGHT + theRect->t) *BLOCK_WIDTH*lightmap_bytes);
			theRect->l = BLOCK_WIDTH;
			theRect->t = BLOCK_HEIGHT;
			theRect->h = 0;
			theRect->w = 0;
		}
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
		glBegin (GL_TRIANGLE_FAN);
		v = p->verts[0];
		for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
		{
			qglMTexCoord2fSGIS (TEXTURE0_SGIS, v[3], v[4]);
			qglMTexCoord2fSGIS (TEXTURE1_SGIS, v[5], v[6]);

			nv[0] = v[0] + 8*sin(v[1]*0.05+realtime)*sin(v[2]*0.05+realtime);
			nv[1] = v[1] + 8*sin(v[0]*0.05+realtime)*sin(v[2]*0.05+realtime);
			nv[2] = v[2];

			glVertex3fv (nv);
		}
		glEnd ();

	} else {
		p = s->polys;

		t = R_TextureAnimation (s->texinfo->texture);
		GL_Bind (t->gl_texturenum);
		DrawGLWaterPoly (p);

		GL_Bind (lightmap_textures + s->lightmaptexturenum);
		glEnable (GL_BLEND);
		DrawGLWaterPolyLightmap (p);
		glDisable (GL_BLEND);
	}
}
#endif


/*
================
DrawGLWaterPoly

Warp the vertex coordinates
================
*/
void DrawGLWaterPoly (glpoly_t *p)
{
	int		i;
	float	*v;
	vec3_t	nv;

	GL_DisableMultitexture();

	glBegin (GL_TRIANGLE_FAN);
	v = p->verts[0];
	for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
	{
		glTexCoord2f (v[3], v[4]);

		nv[0] = v[0] + 8*sin(v[1]*0.05+realtime)*sin(v[2]*0.05+realtime);
		nv[1] = v[1] + 8*sin(v[0]*0.05+realtime)*sin(v[2]*0.05+realtime);
		nv[2] = v[2];

		glVertex3fv (nv);
	}
	glEnd ();
}

void DrawGLWaterPolyLightmap (glpoly_t *p)
{
	int		i;
	float	*v;
	vec3_t	nv;

	GL_DisableMultitexture();

	glBegin (GL_TRIANGLE_FAN);
	v = p->verts[0];
	for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
	{
		glTexCoord2f (v[5], v[6]);

		nv[0] = v[0] + 8*sin(v[1]*0.05+realtime)*sin(v[2]*0.05+realtime);
		nv[1] = v[1] + 8*sin(v[0]*0.05+realtime)*sin(v[2]*0.05+realtime);
		nv[2] = v[2];

		glVertex3fv (nv);
	}
	glEnd ();
}

//rww
static __inline qboolean GL_PointInFrustum(float *point, float radius)
{
	float side;
	int i;
	for (i = 0; i < 6; i++)
	{
		side = g_curFrustum[i][0]*point[0]+
			g_curFrustum[i][1]*point[1]+
			g_curFrustum[i][2]*point[2]+
			g_curFrustum[i][3];

		if (side <= -radius)
		{ //outside of this plane
			return false;
		}
	}

	return true;
}

/*
================
DrawGLPoly
================
*/
//rww - pass in texture, normal mapping path
extern qboolean g_bloomin;
void DrawGLPoly (texture_t *t, int lightmaptex, glpoly_t *p, dlight_t *dlight, entity_t *e, qboolean specPass)
{
	int normalTx = 0;
	if (t && t->gl_normaltexturenum)
	{
		normalTx = t->gl_normaltexturenum;
	}
	else if (dlight)
	{ //if we are rendering for a dlight, force the normal map path
		normalTx = normalmap_default;
	}

	//doesn't seem to often catch things that the node cull misses, don't bother.
	/*
	if (p->cullFrame != r_framecount)
	{
		p->cullFrame = r_framecount;
		if (!e)
		{
			p->culled = (BoxInFrustum(&g_curFrustum, p->mins, p->maxs) == 0);
		}
		else
		{
			p->culled = false;
		}
	}
	if (p->culled)
	{ //frustum-culled
		p->batchPath = true; //don't render the lightmap
		return;
	}
	*/

	if ((g_depthPass && !g_bloomin) || !g_canUseTexFP || !gl_normalMaps.value || !t || normalTx <= 0)
	{
		int		i;
		float	*v;

		if (!g_depthPass)
		{
			p->batchPath = false;
		}

#if 0 //the original method
		glBegin (GL_POLYGON);
		v = p->verts[0];
		for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
		{
			glTexCoord2f (v[3], v[4]);
			glVertex3fv (v);
		}
		glEnd ();
#else //the new method
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), p->batchVerts);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GL_FLOAT), p->batchTC);
		glDrawElements(GL_TRIANGLES, p->numTris*3, GL_UNSIGNED_INT, p->tris);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
#endif
	}
	else
	{ //rww - fancy path
		if (!dlight)
		{
			p->batchPath = true;
		}

		if (lightmaptex >= 0 && lightmap_modified[lightmaptex])
		{ //update the lightmap texture when necessary
			glRect_t *theRect;

			glBindTexture(GL_TEXTURE_2D, lightmap_textures+lightmaptex);
			lightmap_modified[lightmaptex] = false;
			theRect = &lightmap_rectchange[lightmaptex];
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, theRect->t, 
				BLOCK_WIDTH, theRect->h, gl_lightmap_format, GL_UNSIGNED_BYTE,
				lightmaps+(lightmaptex* BLOCK_HEIGHT + theRect->t) *BLOCK_WIDTH*lightmap_bytes);
			theRect->l = BLOCK_WIDTH;
			theRect->t = BLOCK_HEIGHT;
			theRect->h = 0;
			theRect->w = 0;
			glBindTexture(GL_TEXTURE_2D, t->gl_texturenum);
		}

		GL_CalcPolyLightVectors(p, dlight, e, specPass);

		if (specPass)
		{
			g_glExt.glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, g_fpSpecPassH);
			glEnable(GL_FRAGMENT_PROGRAM_ARB);
			g_glExt.glProgramEnvParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 0, gl_specExp.value, 0.0f,
				0.0f, 0.0f);
		}
		else if (!dlight)
		{ //bind the standard lighting program
			if (lightmaptex >= 0)
			{
				g_glExt.glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, g_fpLightAmbientH);
			}
			else
			{ //assume no lightmap means water
				g_glExt.glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, g_fpLightWaterH);
			}
			glEnable(GL_FRAGMENT_PROGRAM_ARB);
		}

		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), p->batchVerts);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GL_FLOAT), p->batchTC);

		//set up the normal map texture
		g_glExt.glActiveTextureARB(GL_TEXTURE1_ARB);
		g_glExt.glClientActiveTextureARB(GL_TEXTURE1_ARB);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, normalTx);
		//stick the light vectors in the tc for this unit, since we can just use normal tc for the nmap
		if (specPass)
		{
			glTexCoordPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), p->haVectors);
		}
		else
		{
			glTexCoordPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), p->lightVectors);
		}
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		if (!dlight && lightmaptex >= 0)
		{ //set up the lightmap texture
			g_glExt.glActiveTextureARB(GL_TEXTURE2_ARB);
			g_glExt.glClientActiveTextureARB(GL_TEXTURE2_ARB);
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, lightmap_textures+lightmaptex);
			glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GL_FLOAT), p->batchTCLM);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		}

		glDrawElements(GL_TRIANGLES, p->numTris*3, GL_UNSIGNED_INT, p->tris);

		if (!dlight && lightmaptex >= 0)
		{
			glDisable(GL_TEXTURE_2D); //disable 2d on texture2
			glDisableClientState(GL_TEXTURE_COORD_ARRAY); //disable tc array on texture2

			g_glExt.glActiveTextureARB(GL_TEXTURE1_ARB);
			g_glExt.glClientActiveTextureARB(GL_TEXTURE1_ARB);
		}

		glDisable(GL_TEXTURE_2D); //disable 2d on texture1
		glDisableClientState(GL_TEXTURE_COORD_ARRAY); //disable tc array on texture1

		g_glExt.glActiveTextureARB(GL_TEXTURE0_ARB);
		g_glExt.glClientActiveTextureARB(GL_TEXTURE0_ARB);

		if (!dlight || specPass)
		{
			glDisable(GL_FRAGMENT_PROGRAM_ARB);
		}

		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
}

//rww
extern float turbsin[];
extern float g_glClearColor[4];
extern int hr_water;
extern int hr_water_n;
extern int hr_water2;
extern int hr_water2_n;
void GL_FancyWaterPoly(msurface_t *fa)
{
	glpoly_t	*p;
	float		*v;
	int			i;
	float		s, t, os, ot;
	float		timescale = 0.5f;

	//blend to these targets based on program params
	g_glExt.glActiveTextureARB(GL_TEXTURE2_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE2_ARB);
	glEnable(GL_TEXTURE_2D);
	//was using hr_water, but it actually looks better to only blend the normal
	//i could optimize the fragment program to reflect this but i'm lazy (and might change
	//my mind again because i'm terrible like that).
	glBindTexture(GL_TEXTURE_2D, hr_water2);
	g_glExt.glActiveTextureARB(GL_TEXTURE3_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE3_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, hr_water_n);
	g_glExt.glActiveTextureARB(GL_TEXTURE0_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE0_ARB);

	//temp hack for a filtering bug
	/*
	GL_Bind(fa->texinfo->texture->gl_normaltexturenum);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	GL_Bind(fa->texinfo->texture->gl_texturenum);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	*/

	g_glExt.glProgramEnvParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 1, 0.5f+(sinf((realtime*timescale))*0.5f), 0.0f,
		0.0f, 0.0f);


	for (p=fa->polys ; p ; p=p->next)
	{
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		for (i=0,v=p->verts[0] ; i<p->numverts ; i++, v+=VERTEXSIZE)
		{
			float *tc = p->batchTC+(i*2);
			float tscale = 0.25f;
			float *vert = p->batchVerts+(i*3);
			float waveOffset;
			os = v[3];
			ot = v[4];

			s = os + (turbsin[(int)((os*0.125+(realtime*timescale)) * (256.0 / (2 * M_PI))) & 255])*tscale;
			s *= (1.0/64);

			t = ot + (turbsin[(int)((ot*0.125+(realtime*timescale)) * (256.0 / (2 * M_PI))) & 255])*tscale;
			t *= (1.0/64);
			tc[0] = s;
			tc[1] = t;
			
			waveOffset = (turbsin[(int)((os*ot*0.125+(realtime*timescale)) * (256.0 / (2 * M_PI))) & 255])*0.5f;
			vert[0] = v[0] + fa->plane->normal[0]*waveOffset;
			vert[1] = v[1] + fa->plane->normal[1]*waveOffset;
			vert[2] = v[2] + fa->plane->normal[2]*waveOffset;
			
		}

		//this is a little wasteful because it double-calculates light vectors. oh well, ha ha ha.
		if (!g_bloomin)
		{
			DrawGLPoly(fa->texinfo->texture, -1, p, NULL, NULL, false);
		}
		else
		{ //dull down for the extra bloom pass
			float f = 0.7f;
			glColor4f(f, f, f, 1.0f);
		}
		if (!g_depthPass || g_bloomin)
		{
			float oldSunPos[3];
			float c[4];
			float sunMove = 128.0f;

			//fudge the sun around a bit
			VectorCopy(g_sunPos, oldSunPos);
			g_sunPos[1] = g_sunPos[0]-g_sunPos[0]*0.25f;
			g_sunPos[2] = -g_sunPos[0];
			g_sunPos[0] += sinf(realtime)*sunMove;
			//g_sunPos[1] += sinf(realtime)*sunMove;
			g_sunPos[2] += cos(realtime)*sunMove;

			c[0] = 0.0f;
			c[1] = 0.0f;
			c[2] = 0.0f;
			c[3] = g_glClearColor[3];
			glFogfv(GL_FOG_COLOR, c);
			glBlendFunc(GL_ONE, GL_ONE);
			DrawGLPoly(fa->texinfo->texture, -1, p, NULL, NULL, true);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glColor4f(1.0f, 1.0f, 1.0f, r_wateralpha.value);
			glFogfv(GL_FOG_COLOR, g_glClearColor);

			VectorCopy(oldSunPos, g_sunPos);
		}

		glColor4f(1.0f, 1.0f, 1.0f, r_wateralpha.value);
	}

	g_glExt.glActiveTextureARB(GL_TEXTURE2_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE2_ARB);
	glDisable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, hr_water2);
	g_glExt.glActiveTextureARB(GL_TEXTURE3_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE3_ARB);
	glDisable(GL_TEXTURE_2D);
	g_glExt.glActiveTextureARB(GL_TEXTURE0_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE0_ARB);
}

/*
================
R_BlendLightmaps
================
*/
void R_BlendLightmaps (void)
{
	int			i, j;
	glpoly_t	*p;
	float		*v;
	glRect_t	*theRect;
	float c[4];

#if 0
	if (r_fullbright.value)
		return;
#endif
	if (!gl_texsort.value)
		return;

	if (g_depthPass)
	{ //rww
		return;
	}

	//rww - change fog for lightmap pass
	c[0] = 0.0f;
	c[1] = 0.0f;
	c[2] = 0.0f;
	c[3] = g_glClearColor[3];
	glFogfv(GL_FOG_COLOR, c);

	glDepthMask (0);		// don't bother writing Z

	if (gl_lightmap_format == GL_LUMINANCE)
		glBlendFunc (GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
	else if (gl_lightmap_format == GL_INTENSITY)
	{
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		glColor4f (0,0,0,1);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	if (!r_lightmap.value)
	{
		glEnable (GL_BLEND);
	}

	for (i=0 ; i<MAX_LIGHTMAPS ; i++)
	{
		p = lightmap_polys[i];
		if (!p)
			continue;

		if (g_canUseTexFP && gl_normalMaps.value && p->batchPath)
		{ //rww
			continue;
		}

		GL_Bind(lightmap_textures+i);
		if (lightmap_modified[i])
		{
			lightmap_modified[i] = false;
			theRect = &lightmap_rectchange[i];
//			theRect->l = 0;
//			theRect->t = 0;
//			theRect->w = BLOCK_WIDTH;
//			theRect->h = BLOCK_HEIGHT;
//			glTexImage2D (GL_TEXTURE_2D, 0, lightmap_bytes
//			, BLOCK_WIDTH, BLOCK_HEIGHT, 0, 
//			gl_lightmap_format, GL_UNSIGNED_BYTE, lightmaps+i*BLOCK_WIDTH*BLOCK_HEIGHT*lightmap_bytes);
//			glTexImage2D (GL_TEXTURE_2D, 0, lightmap_bytes
//				, BLOCK_WIDTH, theRect->h, 0, 
//				gl_lightmap_format, GL_UNSIGNED_BYTE, lightmaps+(i*BLOCK_HEIGHT+theRect->t)*BLOCK_WIDTH*lightmap_bytes);
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, theRect->t, 
				BLOCK_WIDTH, theRect->h, gl_lightmap_format, GL_UNSIGNED_BYTE,
				lightmaps+(i* BLOCK_HEIGHT + theRect->t) *BLOCK_WIDTH*lightmap_bytes);
			theRect->l = BLOCK_WIDTH;
			theRect->t = BLOCK_HEIGHT;
			theRect->h = 0;
			theRect->w = 0;
		}
		for ( ; p ; p=p->chain)
		{
//			if (p->flags & SURF_UNDERWATER)
//				DrawGLWaterPolyLightmap (p);
			if (((r_viewleaf->contents==CONTENTS_EMPTY && (p->flags & SURF_UNDERWATER)) ||
				(r_viewleaf->contents!=CONTENTS_EMPTY && !(p->flags & SURF_UNDERWATER)))
				&& !(p->flags & SURF_DONTWARP) && (!g_canUseTexFP || !gl_normalMaps.value))
				DrawGLWaterPolyLightmap (p);
			else
			{
#if 0
				glBegin (GL_POLYGON);
				v = p->verts[0];
				for (j=0 ; j<p->numverts ; j++, v+= VERTEXSIZE)
				{
					glTexCoord2f (v[5], v[6]);
					glVertex3fv (v);
				}
				glEnd ();
#else //rww
				glEnableClientState(GL_VERTEX_ARRAY);
				glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), p->batchVerts);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GL_FLOAT), p->batchTCLM);
				glDrawElements(GL_TRIANGLES, p->numTris*3, GL_UNSIGNED_INT, p->tris);
				glDisableClientState(GL_VERTEX_ARRAY);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);
#endif
			}
		}
	}

	glDisable (GL_BLEND);
	if (gl_lightmap_format == GL_LUMINANCE)
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	else if (gl_lightmap_format == GL_INTENSITY)
	{
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glColor4f (1,1,1,1);
	}

	glDepthMask (1);		// back to normal Z buffering

	//rww - restore fog color
	glFogfv(GL_FOG_COLOR, g_glClearColor);
}

/*
================
R_RenderBrushPoly
================
*/
void R_RenderBrushPoly (entity_t *e, msurface_t *fa, int lightNum) //rww - ent, light
{
	texture_t	*t;
	byte		*base;
	int			maps;
	glRect_t    *theRect;
	int smax, tmax;
	dlight_t *light = ((lightNum >= 0) ? &cl_dlights[lightNum] : NULL); //rww
	if (light)
	{
		float rad, dist;

		//don't do this for ents, because the dlight push has been modified to only include
		//world nodes which were visible this frame.
		if ( !e && !GL_SurfLitByLight(fa, lightNum) )
			return;		// not lit by this light

		rad = light->radius;
		dist = DotProduct (light->localPos, fa->plane->normal) -
				fa->plane->dist;
		rad -= fabs(dist);
		if (rad < light->minlight)
			return;
	}

	c_brush_polys++;

	if (fa->flags & SURF_DRAWSKY)
	{	// warp texture, no lightmaps
		if (!light)
		{ //rww
			EmitBothSkyLayers (fa);
		}
		return;
	}
		
	t = R_TextureAnimation (fa->texinfo->texture);
	GL_Bind (t->gl_texturenum);

	if (fa->flags & SURF_DRAWTURB)
	{	// warp texture, no lightmaps
		if (!light)
		{ //rww
			EmitWaterPolys (fa);
		}
		return;
	}

	if (((r_viewleaf->contents==CONTENTS_EMPTY && (fa->flags & SURF_UNDERWATER)) ||
		(r_viewleaf->contents!=CONTENTS_EMPTY && !(fa->flags & SURF_UNDERWATER)))
		&& !(fa->flags & SURF_DONTWARP) && (!g_canUseTexFP || !gl_normalMaps.value))
	{ //rww - don't bother with this water stuff for the advanced lighting path
		DrawGLWaterPoly (fa->polys);
	}
	else
	{
		DrawGLPoly (t, fa->lightmaptexturenum, fa->polys, light, e, false);
		if (!light && gl_qualityFoliage.value)
		{ //rww - when not doing light pass, add surface sprites
			if (fa->numSurfSprites > 0 && fa->surfSprites)
			{
				int i;
				for (i = 0; i < fa->numSurfSprites && g_numSurfSpriteDraws < MAX_SURF_SPRITE_DRAWS; i++)
				{
					float t[3];
					surfSpriteDraw_t *d = &g_surfSpriteDraws[g_numSurfSpriteDraws];
					
					d->surf = fa;
					d->spr = fa->surfSprites+i;
					d->mat = fa->surfSpriteMat;
					VectorSubtract(d->spr->basePos, r_refdef.vieworg, t);
					d->viewLen = Length(t);
					if (d->viewLen <= g_beginLodCutoff*0.25f)
					{ //don't add it if it's beyond that surfsprite cutoff
						g_numSurfSpriteDraws++;
					}
				}
			}
		}
	}

	if (light)
	{ //rww
		return;
	}

	// add the poly to the proper lightmap chain

	fa->polys->chain = lightmap_polys[fa->lightmaptexturenum];
	lightmap_polys[fa->lightmaptexturenum] = fa->polys;

	// check for lightmap modification
	for (maps = 0 ; maps < MAXLIGHTMAPS && fa->styles[maps] != 255 ;
		 maps++)
		if (d_lightstylevalue[fa->styles[maps]] != fa->cached_light[maps])
			goto dynamic;

	if ((!g_canUseTexFP || !gl_normalMaps.value) && //rww
		(fa->dlightframe == r_framecount	// dynamic this frame
		|| fa->cached_dlight))			// dynamic previously
	{
dynamic:
		if (r_dynamic.value)
		{
			lightmap_modified[fa->lightmaptexturenum] = true;
			theRect = &lightmap_rectchange[fa->lightmaptexturenum];
			if (fa->light_t < theRect->t) {
				if (theRect->h)
					theRect->h += theRect->t - fa->light_t;
				theRect->t = fa->light_t;
			}
			if (fa->light_s < theRect->l) {
				if (theRect->w)
					theRect->w += theRect->l - fa->light_s;
				theRect->l = fa->light_s;
			}
			smax = (fa->extents[0]>>4)+1;
			tmax = (fa->extents[1]>>4)+1;
			if ((theRect->w + theRect->l) < (fa->light_s + smax))
				theRect->w = (fa->light_s-theRect->l)+smax;
			if ((theRect->h + theRect->t) < (fa->light_t + tmax))
				theRect->h = (fa->light_t-theRect->t)+tmax;
			base = lightmaps + fa->lightmaptexturenum*lightmap_bytes*BLOCK_WIDTH*BLOCK_HEIGHT;
			base += fa->light_t * BLOCK_WIDTH * lightmap_bytes + fa->light_s * lightmap_bytes;
			R_BuildLightMap (fa, base, BLOCK_WIDTH*lightmap_bytes);
		}
	}
}

/*
================
R_RenderDynamicLightmaps
Multitexture
================
*/
void R_RenderDynamicLightmaps (msurface_t *fa)
{
	byte		*base;
	int			maps;
	glRect_t    *theRect;
	int smax, tmax;

	c_brush_polys++;

	if (fa->flags & ( SURF_DRAWSKY | SURF_DRAWTURB) )
		return;
		
	fa->polys->chain = lightmap_polys[fa->lightmaptexturenum];
	lightmap_polys[fa->lightmaptexturenum] = fa->polys;

	// check for lightmap modification
	for (maps = 0 ; maps < MAXLIGHTMAPS && fa->styles[maps] != 255 ;
		 maps++)
		if (d_lightstylevalue[fa->styles[maps]] != fa->cached_light[maps])
			goto dynamic;

	if ((!g_canUseTexFP || !gl_normalMaps.value) && //rww
		(fa->dlightframe == r_framecount	// dynamic this frame
		|| fa->cached_dlight))			// dynamic previously
	{
dynamic:
		if (r_dynamic.value)
		{
			lightmap_modified[fa->lightmaptexturenum] = true;
			theRect = &lightmap_rectchange[fa->lightmaptexturenum];
			if (fa->light_t < theRect->t) {
				if (theRect->h)
					theRect->h += theRect->t - fa->light_t;
				theRect->t = fa->light_t;
			}
			if (fa->light_s < theRect->l) {
				if (theRect->w)
					theRect->w += theRect->l - fa->light_s;
				theRect->l = fa->light_s;
			}
			smax = (fa->extents[0]>>4)+1;
			tmax = (fa->extents[1]>>4)+1;
			if ((theRect->w + theRect->l) < (fa->light_s + smax))
				theRect->w = (fa->light_s-theRect->l)+smax;
			if ((theRect->h + theRect->t) < (fa->light_t + tmax))
				theRect->h = (fa->light_t-theRect->t)+tmax;
			base = lightmaps + fa->lightmaptexturenum*lightmap_bytes*BLOCK_WIDTH*BLOCK_HEIGHT;
			base += fa->light_t * BLOCK_WIDTH * lightmap_bytes + fa->light_s * lightmap_bytes;
			R_BuildLightMap (fa, base, BLOCK_WIDTH*lightmap_bytes);
		}
	}
}

/*
================
R_MirrorChain
================
*/
void R_MirrorChain (msurface_t *s)
{
	if (mirror)
		return;
	mirror = true;
	mirror_plane = s->plane;
}


#if 0
/*
================
R_DrawWaterSurfaces
================
*/
void R_DrawWaterSurfaces (void)
{
	int			i;
	msurface_t	*s;
	texture_t	*t;

	if (r_wateralpha.value == 1.0)
		return;

	//
	// go back to the world matrix
	//
    glLoadMatrixf (r_world_matrix);

	glEnable (GL_BLEND);
	glColor4f (1,1,1,r_wateralpha.value);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	for (i=0 ; i<cl.worldmodel->numtextures ; i++)
	{
		t = cl.worldmodel->textures[i];
		if (!t)
			continue;
		s = t->texturechain;
		if (!s)
			continue;
		if ( !(s->flags & SURF_DRAWTURB) )
			continue;

		// set modulate mode explicitly
		GL_Bind (t->gl_texturenum);

		for ( ; s ; s=s->texturechain)
			R_RenderBrushPoly (NULL, s, -1);

		t->texturechain = NULL;
	}

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glColor4f (1,1,1,1);
	glDisable (GL_BLEND);
}
#else
/*
================
R_DrawWaterSurfaces
================
*/
extern unsigned int g_chaincount; //rww
void R_DrawWaterSurfaces (void)
{
	int			i;
	msurface_t	*s;
	texture_t	*t;

	if (r_wateralpha.value == 1.0 && gl_texsort.value)
		return;

	//
	// go back to the world matrix
	//

    glLoadMatrixf (r_world_matrix);

	if (r_wateralpha.value < 1.0) {
		glEnable (GL_BLEND);
		glColor4f (1,1,1,r_wateralpha.value);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	}

	g_chaincount++; //rww
	if (!gl_texsort.value) {
		if (!waterchain)
			return;

		for ( s = waterchain ; s ; s=s->texturechain) {
			GL_Bind (s->texinfo->texture->gl_texturenum);
			EmitWaterPolys (s);
		}
		
		waterchain = NULL;
	} else {

		for (i=0 ; i<cl.worldmodel->numtextures ; i++)
		{
			t = cl.worldmodel->textures[i];
			if (!t)
				continue;
			s = t->texturechain;
			if (!s)
				continue;
			if ( !(s->flags & SURF_DRAWTURB ) )
				continue;

			// set modulate mode explicitly
			
			GL_Bind (t->gl_texturenum);

			for ( ; s ; s=s->texturechain)
			{
				if (s->chaincount == g_chaincount)
				{ //rww
					break;
				}
				EmitWaterPolys (s);
				s->chaincount = g_chaincount; //rww
			}
			
			//rww - don't do this here, can render multiple times
			//t->texturechain = NULL;
		}

	}

	if (r_wateralpha.value < 1.0) {
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		glColor4f (1,1,1,1);
		glDisable (GL_BLEND);
	}

}

#endif

//rww
static __inline void FastVecNorm(float *v)
{
	if (!g_useSSE)
	{
		float length = sqrtf(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
		if (length)
		{
			length = 1.0f/length;
			v[0] *= length;
			v[1] *= length;
			v[2] *= length;
		}
	}
	else
	{ //the results here aren't exact, only use this function for trivial things (not something like collision)
		if (v[0] == v[1] && v[1] == v[2])
		{
			return;
		}
		__asm
		{
			mov			eax, v

			movlps		xmm0, [eax]		//v[0] and v[1] to xmm0
			movss		xmm1, [eax+8]	//v[2] to xmm1
			movlhps		xmm0, xmm1		//v[2] to high bits with 0 and 1 in xmm0
			movaps		xmm1, xmm0		//store v in xmm1 for later multiplying

			mulps		xmm0, xmm0		//v*v

			//xmm0 = v[0] + v[1] + v[2]
			movaps		xmm2, xmm0
			shufps		xmm0, xmm2, 78
			addps		xmm0, xmm2
			movaps		xmm2, xmm0
			shufps		xmm2, xmm2, 17
			addps		xmm0, xmm2

			//v * 1.0/sqrt
			rsqrtps		xmm0, xmm0
			mulps		xmm1, xmm0

			//move back to v
			movlps		[eax], xmm1
			movhlps		xmm1, xmm1
			movss		[eax+8], xmm1
		}
	}
}

//rww
static __inline void FastVecScale(float *v, float scale)
{
	v[0] *= scale;
	v[1] *= scale;
	v[2] *= scale;
}

//rww
static __inline qboolean GL_TriFacingLight(float *planeEq, float *lightPosition)
{
	float side;

	side = planeEq[0]*lightPosition[0]+
		planeEq[1]*lightPosition[1]+
		planeEq[2]*lightPosition[2]+
		planeEq[3];

	if (side > 0.0f)
	{
		return true;
	}

	return false;
}

//rww
void GL_GenerateShadowVolume(float objRadius, dlight_t *light, qboolean lightInObject, int *checkVerts,
							int *checkTris, int numVerts, int numTris, float *verts, int *tris,
							modelTriNeighbor_t *neighbors, modelVert_t *vertsOut, int *trisOut,
							float *minsOut, float *maxsOut)
{
	float shadowExtrDist;
	float *lightPos;
	int i;
	#define TRI_0		0
	#define TRI_1		1
	#define TRI_2		2
	#define TRIOUT_0	0
	#define TRIOUT_1	1
	#define TRIOUT_2	2

	if (light)
	{
		shadowExtrDist = objRadius+light->radius;
		lightPos = light->localPos;
	}
	else
	{
		shadowExtrDist = objRadius;
		lightPos = currententity->sunLocal;
	}

	for (i = 0; i < numTris; i++)
	{
		//todo - could precalculate face planes, but memory cost is significant for alias models.
		//could at least be done for level geometry.
		float planeEq[4];
		int triVertBase;
		int *tri = tris+(i*3);
		PlaneEquation(verts+(tri[TRI_0]*3), verts+(tri[TRI_1]*3), verts+(tri[TRI_2]*3), planeEq);

		if (lightInObject || GL_TriFacingLight(planeEq, lightPos))
		{ //this face needs to be drawn
			float *v[3];
			float extruded[3][3];
			int *neighbor;
			int j;

			if (lightInObject)
			{ //we'll also be drawing backfaces
				if (GL_TriFacingLight(planeEq, lightPos))
				{
					v[0] = verts+(tri[TRI_0]*3);
					v[1] = verts+(tri[TRI_1]*3);
					v[2] = verts+(tri[TRI_2]*3);
				}
				else
				{
					v[2] = verts+(tri[TRI_0]*3);
					v[1] = verts+(tri[TRI_1]*3);
					v[0] = verts+(tri[TRI_2]*3);
				}
			}
			else
			{
				v[0] = verts+(tri[TRI_0]*3);
				v[1] = verts+(tri[TRI_1]*3);
				v[2] = verts+(tri[TRI_2]*3);
			}

			for (j = 0; j < 3; j++)
			{
				VectorSubtract(v[j], lightPos, extruded[j]);
				FastVecNorm(extruded[j]);
				FastVecScale(extruded[j], shadowExtrDist);
				VectorAdd(v[j], extruded[j], extruded[j]);
			}


			if (minsOut && maxsOut)
			{ //collect mins/maxs while building the volume
				for (j = 0; j < 3; j++)
				{
					int k;
					for (k = 0; k < 3; k++)
					{
						if (v[j][k] < minsOut[k])
						{
							minsOut[k] = v[j][k];
						}
						if (extruded[j][k] < minsOut[k])
						{
							minsOut[k] = extruded[j][k];
						}
						if (v[j][k] > maxsOut[k])
						{
							maxsOut[k] = v[j][k];
						}
						if (extruded[j][k] > maxsOut[k])
						{
							maxsOut[k] = extruded[j][k];
						}
					}
				}
			}

			triVertBase = *checkVerts;
			*checkVerts += 6;
			if (vertsOut)
			{
				vertsOut->x = v[0][0];
				vertsOut->y = v[0][1];
				vertsOut->z = v[0][2];
				vertsOut++;
				vertsOut->x = v[1][0];
				vertsOut->y = v[1][1];
				vertsOut->z = v[1][2];
				vertsOut++;
				vertsOut->x = v[2][0];
				vertsOut->y = v[2][1];
				vertsOut->z = v[2][2];
				vertsOut++;
				vertsOut->x = extruded[0][0];
				vertsOut->y = extruded[0][1];
				vertsOut->z = extruded[0][2];
				vertsOut++;
				vertsOut->x = extruded[1][0];
				vertsOut->y = extruded[1][1];
				vertsOut->z = extruded[1][2];
				vertsOut++;
				vertsOut->x = extruded[2][0];
				vertsOut->y = extruded[2][1];
				vertsOut->z = extruded[2][2];
				vertsOut++;
			}

			neighbor = (int *)(neighbors+i);
			for (j = 0; j < 3; j++)
			{ //see if this strip actually needs to be drawn
				int neighborIndex = neighbor[j];
				qboolean drawStrip = false;

				if (lightInObject)
				{
					drawStrip = true;
				}
				else
				{
					if (neighborIndex != -1)
					{
						float nPlaneEq[4];
						int *nTri = tris+(neighborIndex*3);
						PlaneEquation(verts+(nTri[TRI_0]*3), verts+(nTri[TRI_1]*3), verts+(nTri[TRI_2]*3), nPlaneEq);

						if (!GL_TriFacingLight(nPlaneEq, lightPos))
						{
							drawStrip = true;
						}
					}
					else
					{
						drawStrip = true;
					}
				}

				if (drawStrip)
				{
					int n = j+1;
					if (n > 2)
					{
						n = 0;
					}

					if (trisOut)
					{ //add the strip
						trisOut[TRIOUT_0] = triVertBase+3+n;
						trisOut[TRIOUT_1] = triVertBase+j;
						trisOut[TRIOUT_2] = triVertBase+3+j;
						trisOut += 3;
						trisOut[TRIOUT_0] = triVertBase+n;
						trisOut[TRIOUT_1] = triVertBase+j;
						trisOut[TRIOUT_2] = triVertBase+3+n;
						trisOut += 3;
					}
					*checkTris += 2;
				}
			}

			if (trisOut)
			{ //add front and back caps
				trisOut[TRIOUT_0] = triVertBase;
				trisOut[TRIOUT_1] = triVertBase+1;
				trisOut[TRIOUT_2] = triVertBase+2;
				trisOut += 3;
				trisOut[TRIOUT_0] = triVertBase+5;
				trisOut[TRIOUT_1] = triVertBase+4;
				trisOut[TRIOUT_2] = triVertBase+3;
				trisOut += 3;
			}
			*checkTris += 2;
		}
	}

	if (*checkVerts >= MAX_VOLUME_VERTS || *checkTris >= MAX_VOLUME_TRIS)
	{ //hm!
		Sys_Error("Shadow volume unexpectedly large!");
	}
}

//rww
static __inline int GL_SizeForShadowVolume(int numVerts, int numTris)
{
	return sizeof(shadowVol_t) + (numVerts*(sizeof(float)*3)) + (numTris*(sizeof(int)*3));
}

//rww
static __inline int GL_ExistingVolumeSize(byte *volumes, int numVolumes)
{
	int i;
	int volSize = 0;
	for (i = 0; i < numVolumes; i++)
	{
		shadowVol_t *vol =(shadowVol_t *)volumes;
		volSize += vol->memSize;
		volumes += vol->memSize;
	}

	return volSize;
}

int g_numvolents = 0;
entity_t *g_volents[MAX_VISEDICTS];

//rww
void GL_VolumeEfrags (efrag_t **ppefrag)
{
	entity_t	*pent;
	model_t		*clmodel;
	efrag_t		*pefrag;

	while ((pefrag = *ppefrag) != NULL)
	{
		pent = pefrag->entity;
		clmodel = pent->model;

		switch (clmodel->type)
		{
		case mod_alias:
		case mod_brush:
		case mod_sprite:
		case mod_rdm: //rww
			if (clmodel->type != mod_sprite && g_numvolents < MAX_VISEDICTS &&
				pent->visframe != r_framecount)
			{ //this ent is not in the normal vis list, add it to the volume ents list
				pent = pefrag->entity;
				g_volents[g_numvolents++] = pent;
			}

			ppefrag = &pefrag->leafnext;
			break;

		default:	
			Sys_Error ("R_StoreEfrags: Bad entity type %d\n", clmodel->type);
		}
	}
}

//rww - used as temp buffers for building shadow volumes
float volVerts[3*MAX_VOLUME_VERTS];
int volTris[3*MAX_VOLUME_TRIS];

//rww
void R_WorldNodeVolume(dlight_t *dlight, mnode_t *node)
{
	mplane_t	*splitplane;
	float		dist;
	msurface_t	*surf;
	int			i;
	
	if (node->contents < 0)
	{
		mleaf_t *pleaf = (mleaf_t *)node;
		if (pleaf->efrags)
		{ //add efrags in this leaf for potential extra ent obstructors
			GL_VolumeEfrags(&pleaf->efrags);
		}
		return;
	}

	splitplane = node->plane;
	dist = DotProduct (dlight->origin, splitplane->normal) - splitplane->dist;
	
	if (dist > dlight->radius)
	{
		R_WorldNodeVolume (dlight, node->children[0]);
		return;
	}
	if (dist < -dlight->radius)
	{
		R_WorldNodeVolume (dlight, node->children[1]);
		return;
	}
		
	surf = cl.worldmodel->surfaces + node->firstsurface;
	for (i=0 ; i<node->numsurfaces ; i++, surf++)
	{
		float volMins[3], volMaxs[3];
		int checkVerts, checkTris;

		checkVerts = 0;
		checkTris = 0;
		volMins[0] = surf->polys->batchVerts[0]-0.01f;
		volMins[1] = surf->polys->batchVerts[1]-0.01f;
		volMins[2] = surf->polys->batchVerts[2]-0.01f;
		volMaxs[0] = surf->polys->batchVerts[0]+0.01f;
		volMaxs[1] = surf->polys->batchVerts[1]+0.01f;
		volMaxs[2] = surf->polys->batchVerts[2]+0.01f;
		GL_GenerateShadowVolume(surf->polys->radius, dlight, false, &checkVerts, &checkTris,
			surf->polys->numverts, surf->polys->numTris, surf->polys->batchVerts, surf->polys->tris,
			surf->polys->neighbors, (modelVert_t *)volVerts, volTris, volMins, volMaxs);
		if (checkVerts > 0 && checkTris > 0)
		{
			if (dlight->nonmoving)
			{ //need to build a cache
				shadowVol_t *vol;
				int curSize = GL_ExistingVolumeSize(dlight->worldVolumes, dlight->numVolumes);
				int volSize = GL_SizeForShadowVolume(checkVerts, checkTris);
				int j;
				BYTE *b;

				dlight->worldVolumes = (byte *)realloc(dlight->worldVolumes, curSize+volSize);
				b = dlight->worldVolumes;
				for (j = 0; j < dlight->numVolumes; j++)
				{
					vol = (shadowVol_t *)b;
					vol->verts = (float *)(b+sizeof(shadowVol_t));
					vol->tris = (int *)(vol->verts+(3*vol->numVerts));
					b += vol->memSize;
				}
				vol = (shadowVol_t *)b;
				memset(vol, 0, volSize);
				vol->memSize = volSize;
				vol->numVerts = checkVerts;
				vol->numTris = checkTris;
				vol->verts = (float *)(b+sizeof(shadowVol_t));
				vol->tris = (int *)(vol->verts+(3*vol->numVerts));
				memcpy(vol->verts, volVerts, sizeof(float)*3*vol->numVerts);
				memcpy(vol->tris, volTris, sizeof(int)*3*vol->numTris);
				VectorCopy(volMins, vol->mins);
				VectorCopy(volMaxs, vol->maxs);
				dlight->numVolumes++;
			}

			if (!R_CullBox(volMins, volMaxs))
			{ //no need to draw it if it's culled from view
				glEnableClientState(GL_VERTEX_ARRAY);
				glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), volVerts);
				glDrawElements(GL_TRIANGLES, checkTris*3, GL_UNSIGNED_INT, volTris);
				glDisableClientState(GL_VERTEX_ARRAY);
			}
		}
	}

	R_WorldNodeVolume (dlight, node->children[0]);
	R_WorldNodeVolume (dlight, node->children[1]);
}

//rww
void GL_SetupCurEntityForDLight(dlight_t *dlight)
{
	if (currententity->modelMatFrame != r_framecount)
	{ //update the entity transform matrices
		currententity->modelMatFrame = r_framecount;
		glPushMatrix();
		glLoadIdentity();
		glTranslatef (currententity->origin[0],  currententity->origin[1],  currententity->origin[2]);
		glRotatef (currententity->angles[1],  0, 0, 1);
		glRotatef (-currententity->angles[0],  0, 1, 0);
		glRotatef (currententity->angles[2],  1, 0, 0);
		if (currententity->model->type == mod_alias)
		{ //stupid hack for existing stupid hack that's used for alias models
			model_t *clmodel = currententity->model;
			aliashdr_t *paliashdr = (aliashdr_t *)Mod_Extradata (clmodel);

			/*if (!strcmp (clmodel->name, "progs/eyes.mdl") ) {
				glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2] - (22 + 8));
			// double size of eyes, since they are really hard to see in gl
				glScalef (paliashdr->scale[0]*2, paliashdr->scale[1]*2, paliashdr->scale[2]*2);
			} else*/ { //rww - don't care about eye size hack
				glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
				glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);
			}
		}
		glGetFloatv(GL_MODELVIEW_MATRIX, currententity->modelMat);
		glPopMatrix();
		MatrixInverse4x4(currententity->modelMat, currententity->invModelMat);

		if (g_hasSun)
		{ //transform the sun pos to local for sun rendering passes
			TransformPointByMatrix4x4(currententity->invModelMat, g_sunPos, currententity->sunLocal);
		}

		//todo - implement viewLocal for things which need it (currently no ents do)
	}

	if (dlight && (dlight->localEnt != currententity || dlight->localEntFrame != r_framecount))
	{ //need to get the local position for this ent
		dlight->localEntFrame = r_framecount;
		TransformPointByMatrix4x4(currententity->invModelMat, dlight->origin, dlight->localPos);
		dlight->localEnt = currententity;
	}
}

//rww
qboolean GL_GlowModelName(const char *str);
void GL_DrawAliasVolume(entity_t *e, int lightNum)
{
	model_t		*clmodel;
	aliashdr_t	*paliashdr;
	float volMins[3], volMaxs[3];
	int checkVerts, checkTris;
	int pose, numposes;
	int frame = e->frame;
	float interval;
	int i;
	dlight_t *dlight = NULL;
	const float fudge = 1.0f;
	qboolean inLight = true;
	float aliasscale = 1.0f;
	float mins[3], maxs[3];

	if (lightNum >= 0)
	{
		dlight = &cl_dlights[lightNum];
	}

	clmodel = e->model;

	if (GL_GlowModelName(clmodel->name))
	{ //don't draw volumes for torches and such
		return;
	}

	paliashdr = (aliashdr_t *)Mod_Extradata (e->model);

	VectorAdd (e->origin, clmodel->mins, mins);
	VectorAdd (e->origin, clmodel->maxs, maxs);
	if (dlight)
	{
		for (i = 0; i < 3; i++)
		{
			if (dlight->origin[i]-fudge > maxs[i])
			{
				inLight = false;
				break;
			}
			if (dlight->origin[i]+fudge < mins[i])
			{
				inLight = false;
				break;
			}
		}
	}
	else
	{
		inLight = false;
	}

    glPushMatrix ();
	R_RotateForEntity (e);

	/*if (!strcmp (clmodel->name, "progs/eyes.mdl") ) {
		glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2] - (22 + 8));
	// double size of eyes, since they are really hard to see in gl
		glScalef (paliashdr->scale[0]*2, paliashdr->scale[1]*2, paliashdr->scale[2]*2);
	} else*/ { //rww - don't care about eye size hack
		glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
		glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);
	}

	if ((frame >= paliashdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
		frame = 0;
	}
	pose = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;
	if (numposes > 1)
	{
		interval = paliashdr->frames[frame].interval;
		pose += (int)(cl.time / interval) % numposes;
	}

	checkVerts = 0;
	checkTris = 0;
	volMins[0] = paliashdr->batchVerts[0]-0.01f;
	volMins[1] = paliashdr->batchVerts[1]-0.01f;
	volMins[2] = paliashdr->batchVerts[2]-0.01f;
	volMaxs[0] = paliashdr->batchVerts[0]+0.01f;
	volMaxs[1] = paliashdr->batchVerts[1]+0.01f;
	volMaxs[2] = paliashdr->batchVerts[2]+0.01f;
	for (i = 0; i < 3; i++)
	{ //scale radius for the volume projection for alias models
		if (paliashdr->scale[i] < aliasscale)
		{
			aliasscale = paliashdr->scale[i];
		}
	}
	GL_GenerateShadowVolume(paliashdr->radius*(1.0f/aliasscale), dlight, inLight, &checkVerts, &checkTris,
		paliashdr->numverts, paliashdr->numtris, paliashdr->batchVerts+(pose*paliashdr->numverts*3), paliashdr->tris,
		paliashdr->neighbors, (modelVert_t *)volVerts, volTris, volMins, volMaxs);
	if (checkVerts > 0 && checkTris > 0)
	{
		for (i = 0; i < 3; i++)
		{
			volMins[i] += paliashdr->scale_origin[i];
			volMins[i] *= paliashdr->scale[i];
			volMaxs[i] += paliashdr->scale_origin[i];
			volMaxs[i] *= paliashdr->scale[i];
		}
		VectorAdd(e->origin, volMins, volMins);
		VectorAdd(e->origin, volMaxs, volMaxs);

		if (!R_CullBox(volMins, volMaxs))
		{ //no need to draw it if it's culled from view
			glEnableClientState(GL_VERTEX_ARRAY);
			glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), volVerts);
			glDrawElements(GL_TRIANGLES, checkTris*3, GL_UNSIGNED_INT, volTris);
			glDisableClientState(GL_VERTEX_ARRAY);
		}
	}

	glPopMatrix();
}

//rww
void GL_DrawBrushModelVolume(entity_t *e, int lightNum)
{
	msurface_t	*psurf;
	float		dot;
	mplane_t	*pplane;
	model_t		*clmodel;
	int			i;
	dlight_t	*dlight = NULL;
	float		*lightPos;
	if (lightNum >= 0)
	{
		dlight = &cl_dlights[lightNum];
		lightPos = dlight->origin;
	}
	else
	{
		lightPos = g_sunPos;
	}

	clmodel = e->model;

	/*
	VectorSubtract (lightPos, e->origin, modelorg);
	if (e->angles[0] || e->angles[1] || e->angles[2])
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (modelorg, temp);
		AngleVectors (e->angles, forward, right, up);
		modelorg[0] = DotProduct (temp, forward);
		modelorg[1] = -DotProduct (temp, right);
		modelorg[2] = DotProduct (temp, up);
	}
	*/


    glPushMatrix ();
e->angles[0] = -e->angles[0];	// stupid quake bug
	R_RotateForEntity (e);
e->angles[0] = -e->angles[0];	// stupid quake bug

	psurf = &clmodel->surfaces[clmodel->firstmodelsurface];
	for (i=0 ; i<clmodel->nummodelsurfaces ; i++, psurf++)
	{
	// find which side of the node we are on
		//this doesn't appear to work out for volumes, fixme
		/*
		pplane = psurf->plane;

		dot = DotProduct (modelorg, pplane->normal) - pplane->dist;

	// draw the polygon
		if (!dlight || ((psurf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) ||
			(!(psurf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))*/
		{
			float volMins[3], volMaxs[3];
			int checkVerts, checkTris;
			qboolean inLight = true;

			if (!dlight)
			{
				inLight = false;
			}

			checkVerts = 0;
			checkTris = 0;
			volMins[0] = psurf->polys->batchVerts[0]-0.01f;
			volMins[1] = psurf->polys->batchVerts[1]-0.01f;
			volMins[2] = psurf->polys->batchVerts[2]-0.01f;
			volMaxs[0] = psurf->polys->batchVerts[0]+0.01f;
			volMaxs[1] = psurf->polys->batchVerts[1]+0.01f;
			volMaxs[2] = psurf->polys->batchVerts[2]+0.01f;
			//fixme - doing two-sided crap for these because the planes are jacked up
			GL_GenerateShadowVolume(psurf->polys->radius*1.5f, dlight, inLight, &checkVerts, &checkTris,
				psurf->polys->numverts, psurf->polys->numTris, psurf->polys->batchVerts, psurf->polys->tris,
				psurf->polys->neighbors, (modelVert_t *)volVerts, volTris, volMins, volMaxs);
			if (checkVerts > 0 && checkTris > 0)
			{
				VectorAdd(e->origin, volMins, volMins);
				VectorAdd(e->origin, volMaxs, volMaxs);
				if (!R_CullBox(volMins, volMaxs))
				{ //no need to draw it if it's culled from view
					glEnableClientState(GL_VERTEX_ARRAY);
					glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), volVerts);
					glDrawElements(GL_TRIANGLES, checkTris*3, GL_UNSIGNED_INT, volTris);
					glDisableClientState(GL_VERTEX_ARRAY);
				}
			}
		}
	}

	glPopMatrix ();
}

//rww - draw shadow volumes
int GL_DrawVolumesForDLight(int chaincount, int lightNum)
{
	int i;
	dlight_t *dlight = &cl_dlights[lightNum];

	dlight->localEnt = NULL;
	dlight->localPos[0] = dlight->origin[0];
	dlight->localPos[1] = dlight->origin[1];
	dlight->localPos[2] = dlight->origin[2];

	g_numvolents = 0;

	//draw the world's volumes
	if (dlight->visframe == r_framecount)
	{
		if (!dlight->nonmoving || !dlight->hasworldgeo)
		{ //dynamic, or need to build geometry
			R_WorldNodeVolume(dlight, cl.worldmodel->nodes);
			if (dlight->nonmoving)
			{
				dlight->hasworldgeo = true;
			}
		}
		else if (dlight->worldVolumes)
		{ //already has a cache of volumes for world geometry
			BYTE *b = dlight->worldVolumes;
			int j;
			for (j = 0; j < dlight->numVolumes; j++)
			{
				shadowVol_t *vol = (shadowVol_t *)b;
				if (!R_CullBox(vol->mins, vol->maxs))
				{ //no need to draw it if it's culled from view
					glEnableClientState(GL_VERTEX_ARRAY);
					glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), vol->verts);
					glDrawElements(GL_TRIANGLES, vol->numTris*3, GL_UNSIGNED_INT, vol->tris);
					glDisableClientState(GL_VERTEX_ARRAY);
				}
				b += vol->memSize;
			}
		}
	}

	//draw entity volumes
	for (i=0 ; i<cl_numvisedicts+g_numvolents ; i++)
	{
		if (i >= cl_numvisedicts)
		{
			currententity = g_volents[i-cl_numvisedicts];
		}
		else
		{
			currententity = &cl_visedicts[i];
		}

		if (currententity->noShadows)
		{ //skip shadows for this ent
			continue;
		}

		if (currententity->model->type == mod_alias ||
			currententity->model->type == mod_brush ||
			currententity->model->type == mod_rdm)
		{
			if (currententity->model->type == mod_alias)
			{
				if (GL_AliasModelInLight(currententity, dlight))
				{
					GL_SetupCurEntityForDLight(dlight);
					GL_DrawAliasVolume(currententity, lightNum);
				}
			}
			else if (currententity->model->type == mod_rdm)
			{
				if (GL_RDMModelInLight(currententity, dlight))
				{
					GL_SetupCurEntityForDLight(dlight);
					GL_DrawRDMVolume(currententity, lightNum);
				}
			}
			else
			{
				if (GL_BrushModelInLight(currententity, dlight))
				{
					GL_SetupCurEntityForDLight(dlight);
					GL_DrawBrushModelVolume(currententity, lightNum);
				}
			}
		}
	}

	return chaincount;
}

//rww - do the stencil pass for a light
int GL_DrawDLightStencil(int chaincount, int lightNum)
{
	const int stenInRef = 0;
	const int stenRef = 0;
	const int stenMask = 0xFFFFFFFF;

	glDisable(GL_TEXTURE_2D);

	glEnable(GL_STENCIL_TEST);

#if 0
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);
	glColor4f(0.05f, 0.05f, 0.05f, 1.0f);
#else
	glColorMask( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );
#endif

	glStencilFunc(GL_ALWAYS, stenRef, stenMask);
	glDepthFunc(GL_LESS);
	glEnable(GL_CULL_FACE);

	if (g_twoSidedStencil && gl_twoSidedStencil.value)
	{
		glEnable(GL_STENCIL_TEST_TWO_SIDE_EXT);

		g_glExt.glActiveStencilFaceEXT(GL_BACK);
		glStencilOp(GL_KEEP, GL_DECR_WRAP_EXT, GL_KEEP);

		g_glExt.glActiveStencilFaceEXT(GL_FRONT);
		glStencilOp(GL_KEEP, GL_INCR_WRAP_EXT, GL_KEEP);

		glDisable(GL_CULL_FACE);

		chaincount = GL_DrawVolumesForDLight(chaincount, lightNum);

		glDisable(GL_STENCIL_TEST_TWO_SIDE_EXT);
		glEnable(GL_CULL_FACE);
	}
	else
	{
		//first pass
		glCullFace(GL_FRONT);
		glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
		chaincount = GL_DrawVolumesForDLight(chaincount, lightNum);

		//second pass
		glCullFace(GL_BACK);
		glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);
		chaincount = GL_DrawVolumesForDLight(chaincount, lightNum);
	}

	glCullFace(GL_FRONT);
	glDepthFunc(GL_LEQUAL);

	//set the mask up for the subsequent lighting pass
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glStencilFunc(GL_EQUAL, stenInRef, stenMask);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

#if 0
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
#endif

	glEnable(GL_TEXTURE_2D);

	return chaincount;
}

//rww
qboolean GL_RDMModelInLight(entity_t *e, dlight_t *dlight)
{
	modelFile_t *model = Model_CoreFromQuakeEnt(e);
	float mins[3], maxs[3], lmins[3], lmaxs[3];
	int i;
	float baseScale;

	if (!model || !model->firstObject)
	{
		return false;
	}

	baseScale = model->baseScale;
	if (!baseScale)
	{
		baseScale = 1.0f;
	}

	lmins[0] = dlight->origin[0]-dlight->radius;
	lmins[1] = dlight->origin[1]-dlight->radius;
	lmins[2] = dlight->origin[2]-dlight->radius;
	lmaxs[0] = dlight->origin[0]+dlight->radius;
	lmaxs[1] = dlight->origin[1]+dlight->radius;
	lmaxs[2] = dlight->origin[2]+dlight->radius;
	mins[0] = e->origin[0]-(e->model->radius*baseScale);
	mins[1] = e->origin[1]-(e->model->radius*baseScale);
	mins[2] = e->origin[2]-(e->model->radius*baseScale);
	maxs[0] = e->origin[0]+(e->model->radius*baseScale);
	maxs[1] = e->origin[1]+(e->model->radius*baseScale);
	maxs[2] = e->origin[2]+(e->model->radius*baseScale);

	if (BoxesOverlap(lmins, lmaxs, mins, maxs))
	{
		return true;
	}

	return false;
}

//rww
qboolean GL_AliasModelInLight(entity_t *e, dlight_t *dlight)
{
	float d[3];
	aliashdr_t *paliashdr;
	model_t *model = e->model;

	if (!model || model->type != mod_alias)
	{
		return false;
	}
	paliashdr = (aliashdr_t *)Mod_Extradata(model);
	if (!paliashdr)
	{
		return false;
	}

	VectorSubtract(e->origin, dlight->origin, d);
	if (Length(d) <= (paliashdr->radius+dlight->radius))
	{
		return true;
	}

	return false;
}

//rww
qboolean GL_BrushModelInLight(entity_t *e, dlight_t *dlight)
{
	model_t *clmodel = e->model;
	int i;
	float mins[3], maxs[3];
	float lmins[3] = {dlight->origin[0]-dlight->radius, dlight->origin[1]-dlight->radius, dlight->origin[2]-dlight->radius};
	float lmaxs[3] = {dlight->origin[0]+dlight->radius, dlight->origin[1]+dlight->radius, dlight->origin[2]+dlight->radius};

	if (e->angles[0] || e->angles[1] || e->angles[2])
	{
		for (i=0 ; i<3 ; i++)
		{
			mins[i] = e->origin[i] - clmodel->radius;
			maxs[i] = e->origin[i] + clmodel->radius;
		}
	}
	else
	{
		VectorAdd (e->origin, clmodel->mins, mins);
		VectorAdd (e->origin, clmodel->maxs, maxs);
	}

	if (BoxesOverlap(mins, maxs, lmins, lmaxs))
	{
		return true;
	}

	return false;
}

extern int window_width, window_height;
#define SCIS_W window_width //vid.width
#define SCIS_H window_height //vid.height

//rww
//chop function needs to be optimized. need to use output for depth bounds extension.
qboolean GL_ScissorLight(dlight_t *light, float *scis)
{
	#define MAX_SCISSOR_POINTS 128
	float lightMins[3], lightMaxs[3];
	float screenPts[MAX_SCISSOR_POINTS][3];
	float boundsPoints[MAX_SCISSOR_POINTS][3];
	float vMin[3], vMax[3];
	float bigW = 0.0f, bigH = 0.0f;
	float minZ = 1.0f;
	float maxZ = 0.0f;
	int i, m;
	int numBoundsPoints, boundsFrust;

	const float fudge = 0.0f;//32.0f;
	for (m = 0; m < 3; m++)
	{
		lightMins[m] = light->origin[m]-(light->radius+fudge);
		lightMaxs[m] = light->origin[m]+(light->radius+fudge);
		vMin[m] = r_refdef.vieworg[m]-64.0f;
		vMax[m] = r_refdef.vieworg[m]+64.0f;
	}
	if (BoxesOverlap(lightMins, lightMaxs, vMin, vMax))
	{ //overlaps with view
		scis[0] = 0.0f;
		scis[1] = 0.0f;
		scis[2] = (float)SCIS_W;
		scis[3] = (float)SCIS_H;
		scis[4] = 0.0f;
		scis[5] = 1.0f;
		return true;
	}

	//intersect the frustum with the bounds
	boundsFrust = BoxInFrustum(&g_curFrustum, lightMins, lightMaxs);
	if (!boundsFrust)
	{ //box outside frustum (shouldn't happen after radius check)
		return false;
	}

	if (boundsFrust == 1)
	{ //box is completely inside the frustum
		numBoundsPoints = 8;
		GetBoundsPoints(lightMins, lightMaxs, (float *)boundsPoints, NULL);
	}
	else
	{ //box is intersecting the frustum
		numBoundsPoints = ChopBoxToFrustum(&g_curFrustum, lightMins, lightMaxs,
			(float *)boundsPoints, MAX_SCISSOR_POINTS, 1.0f);//0.5f);
		if (numBoundsPoints < 1)
		{ //should not happen at this point
			return false;
		}
	}

	//convert each world point into screenspace using the current projection matrix
	for (i = 0; i < numBoundsPoints; i++)
	{
		float pSpace[4];
		float modelMat[16];
		float d[3];

		memcpy(modelMat, r_world_matrix, sizeof(modelMat));
		TranslateMatrix4x4(modelMat, boundsPoints[i]);

		TransformV4ByMatrix4x4(g_projectionMatrix, &modelMat[12], pSpace);
		if (!pSpace[3])
		{
			d[0] = 0.0f;
			d[1] = 0.0f;
			d[2] = 0.0f;
		}
		else
		{
			d[0] = pSpace[0]/pSpace[3];
			d[1] = pSpace[1]/pSpace[3];
			d[2] = pSpace[2]/pSpace[3];
		}
		screenPts[i][0] = ((float)SCIS_W*(0.5f)*(d[0]+1.0f));
		screenPts[i][1] = (float)SCIS_H - ((float)SCIS_H*(0.5f)*(d[1]+1.0f));
		screenPts[i][2] = ((0.5f)*(d[2]+1.0f));

		if (screenPts[i][2] < minZ)
		{
			minZ = screenPts[i][2];
		}
		if (screenPts[i][2] > maxZ)
		{
			maxZ = screenPts[i][2];
		}
	}

	if (minZ < 0.0f)
	{
		minZ = 0.0f;
	}
	if (maxZ > 1.0f)
	{
		maxZ = 1.0f;
	}
	scis[0] = (float)SCIS_W;
	scis[1] = (float)SCIS_H;
	scis[2] = 0.0f;
	scis[3] = 0.0f;
	scis[4] = minZ;
	scis[5] = maxZ;

	for (i = 0; i < numBoundsPoints; i++)
	{
		if (screenPts[i][0] < scis[0])
		{
			scis[0] = screenPts[i][0];
		}
		if (screenPts[i][1] < scis[1])
		{
			scis[1] = screenPts[i][1];
		}

		if (screenPts[i][0] > bigW)
		{
			bigW = screenPts[i][0];
		}
		if (screenPts[i][1] > bigH)
		{
			bigH = screenPts[i][1];
		}
	}

	if (scis[0] < 0.0f)
	{
		scis[0] = 0.0f;
	}
	else if (scis[0] > (float)SCIS_W)
	{
		scis[0] = (float)SCIS_W;
	}
	if (scis[1] < 0.0f)
	{
		scis[1] = 0.0f;
	}
	else if (scis[1] > (float)SCIS_H)
	{
		scis[1] = (float)SCIS_H;
	}

	if (bigW > (float)SCIS_W)
	{
		bigW = (float)SCIS_W;
	}
	if (bigH > (float)SCIS_H)
	{
		bigH = (float)SCIS_H;
	}

	scis[2] = bigW-scis[0];
	scis[3] = bigH-scis[1];

	return true;
}

//rww
float GL_FrustumLightCull(float *point, float radius)
{
	float side;
	int i = 0;
	float r = 0.0f;
	while (i < 6)
	{
		side = g_curFrustum[i][0]*point[0]+
			g_curFrustum[i][1]*point[1]+
			g_curFrustum[i][2]*point[2]+
			g_curFrustum[i][3];

		if (side <= -radius)
		{ //outside of this plane
			return 1.1f;
		}

		if (side < 0.0f)
		{
			float d = (-side)/radius;
			if (d > r)
			{
				r = d;
			}
		}

		i++;
	}

	r = 1.0f-r;
	if (r < 0.0f)
	{
		r = 0.0f;
	}
	else if (r > 1.0f)
	{
		r = 1.0f;
	}
	return r;
}

//rww - draw all surfaces that interact with a dlight using the normal mapping path
extern float g_lightExposureFactor;
extern cvar_t gl_lightScissor;
qboolean GL_DrawDLightSurfs(int chaincount, int lightNum)
{
	dlight_t *dlight = &cl_dlights[lightNum];
	int i;
	msurface_t *s;
	texture_t *t;
	float rad;
	float d[3];
	float lightDistToView;

	//set up gl state for the light
	const float arbitraryFactor = 0.90f;
	const float PS[4] = {1.0f, 0.0f, 0.0f, 0.0f};
	const float PT[4] = {0.0f, 1.0f, 0.0f, 0.0f};
	const float PR[4] = {0.0f, 0.0f, 1.0f, 0.0f};
	const float PQ[4] = {0.0f, 0.0f, 0.0f, 1.0f};
	float clrFactor = 1.0f;
	float clr[4];
	float lmins[3] = {dlight->origin[0]-dlight->radius, dlight->origin[1]-dlight->radius, dlight->origin[2]-dlight->radius};
	float lmaxs[3] = {dlight->origin[0]+dlight->radius, dlight->origin[1]+dlight->radius, dlight->origin[2]+dlight->radius};

	if (R_CullBox(lmins, lmaxs))
	{
		return false;
	}

	//may want to use this later for culling out shadows and such for super-long distances
	VectorSubtract(r_refdef.vieworg, dlight->origin, d);
	lightDistToView = Length(d);
	if (lightDistToView <= 0.001f)
	{
		lightDistToView = 0.001f;
	}

	if (lightDistToView >= g_endLodCutoff)
	{ //clipped off
		return false;
	}
	else if (lightDistToView >= g_beginLodCutoff)
	{ //fading
		clrFactor = 1.0f-((lightDistToView-g_beginLodCutoff)/(g_endLodCutoff-g_beginLodCutoff));
	}

	if (lightDistToView < (dlight->radius*1.5f))
	{
		float lightFactor;
		float fwd[3];
		float frustRad;
		/*
		float frustRad = GL_FrustumLightCull(dlight->origin, dlight->radius);
		if (frustRad > 1.0f)
		{ //outside completely
			return;
		}
		*/

		//basis for light bloom
		if (dlight->nonmoving)
		{
			lightFactor = dlight->radius/1024.0f;
		}
		else
		{
			lightFactor = dlight->radius/768.0f;
		}

		/*
		frustRad *= 0.5f;
		if (frustRad >= 0.5f)
		{
			float fwd[3];
			float dot;
			AngleVectors(r_refdef.viewangles, fwd, 0, 0);
			VectorNormalize(d);
			dot = DotProduct(fwd, d);
			dot += 0.7f;
			dot = ((-dot)*1.6f);
			if (dot > 0.5f)
			{
				dot = 0.5f;
			}
			else if (dot < 0.0f)
			{
				dot = 0.0f;
			}
			frustRad += dot;
		}
		*/

		AngleVectors(r_refdef.viewangles, fwd, 0, 0);
		VectorNormalize(d);
		frustRad = DotProduct(fwd, d);

		if (frustRad < 0.0f)
		{
			frustRad = -frustRad;
			g_lightExposureFactor += (frustRad*(1.0f-(lightDistToView/(dlight->radius*1.5f))))*lightFactor;
		}
	}


	//debugging volume bounds
	/*
	if (dlight->worldVolumes)
	{
		BYTE *b = dlight->worldVolumes;
		int j;
		for (j = 0; j < dlight->numVolumes; j++)
		{
			shadowVol_t *vol = (shadowVol_t *)b;
			glDisable(GL_DEPTH_TEST);
			GL_DebugDrawLineMinsMaxs(vol->mins, vol->maxs, NULL);
			glEnable(GL_DEPTH_TEST);
			b += vol->memSize;
		}
	}
	*/

	if (gl_lightScissor.value)
	{
		if (dlight->scissorFrame != r_framecount)
		{ //create the scissor rect for the light
			dlight->scissorFrame = r_framecount;
			dlight->doScissor = GL_ScissorLight(dlight, dlight->frameScissor);
		}

		if (!dlight->doScissor)
		{ //if this is the case, it was completely clipped off during scissor
			return false;
		}

		//set the scissor
		glScissor((int)dlight->frameScissor[0], (SCIS_H-(int)dlight->frameScissor[1]) - (int)dlight->frameScissor[3], (int)dlight->frameScissor[2], (int)dlight->frameScissor[3]);
		glEnable(GL_SCISSOR_TEST);
	}

	//draw the stencil mask
	chaincount = GL_DrawDLightStencil(chaincount, lightNum);

	g_glExt.glActiveTextureARB(GL_TEXTURE2_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE2_ARB);

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);
	g_glExt.glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, g_fpLightAmbientDLightH);
	glEnable(GL_FRAGMENT_PROGRAM_ARB);
	for (i = 0; i < 3; i++)
	{
		clr[i] = dlight->color[i]*clrFactor;
		if (clr[i] > 1.0f)
		{
			clr[i] = 1.0f;
		}
	}
	clr[3] = 1.0f;
	glColor4fv(clr);
	//set up the attenuation texture pass
	g_glExt.glActiveTextureARB(GL_TEXTURE2_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE2_ARB);
	glEnable(GL_TEXTURE_3D);
	glBindTexture(GL_TEXTURE_3D, attenmap_default);

	//set the texture matrix
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glTranslatef(0.5f, 0.5f, 0.5f);
	glScalef(0.5f, 0.5f, 0.5f);
	rad = dlight->radius;
	if (dlight->flickerEff)
	{ //hackhackhackhackhack
		rad -= ((float)(rand()%8192)/8192.0f)*4.0f;
	}
	glScalef(1.0f/(rad*arbitraryFactor), 1.0f/(rad*arbitraryFactor), 1.0f/(rad*arbitraryFactor));
	
	glTranslatef(-dlight->origin[0], -dlight->origin[1], -dlight->origin[2]);

	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glEnable(GL_TEXTURE_GEN_R);
	glEnable(GL_TEXTURE_GEN_Q);

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGenfv(GL_S, GL_EYE_PLANE, PS);
	glTexGenfv(GL_T, GL_EYE_PLANE, PT);
	glTexGenfv(GL_R, GL_EYE_PLANE, PR);
	glTexGenfv(GL_Q, GL_EYE_PLANE, PQ);

	glMatrixMode(GL_MODELVIEW);

	g_glExt.glActiveTextureARB(GL_TEXTURE0_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE0_ARB);

	//draw the world
	if (dlight->visframe == r_framecount)
	{
		for (i=0 ; i<cl.worldmodel->numtextures ; i++)
		{
			t = cl.worldmodel->textures[i];
			if (!t)
				continue;
			s = t->texturechain;
			if (!s)
				continue;
			if (i == skytexturenum)
				continue;
			else if (i == mirrortexturenum && r_mirroralpha.value != 1.0)
			{
				continue;
			}
			else
			{
				if ((s->flags & SURF_DRAWTURB) && r_wateralpha.value != 1.0)
					continue;	// draw translucent water later
				if ((s->flags & SURF_DRAWTURB) || (s->flags & SURF_DRAWSKY))
				{ //don't light water/sky
					continue;
				}
				for ( ; s ; s=s->texturechain)
				{
					float rad, dist;
					texture_t *t;

					if (s->chaincount == chaincount)
					{
						break;
					}
					s->chaincount = chaincount;

					/*
					if (s->visframe != r_framecount)
					{
						continue;
					}
					*/
					//(not necessary)

					if ( !GL_SurfLitByLight(s, lightNum) )
						continue;		// not lit by this light

					rad = dlight->radius;
					dist = DotProduct (dlight->origin, s->plane->normal) -
							s->plane->dist;
					rad -= fabs(dist);
					if (rad < dlight->minlight)
						continue;

					t = R_TextureAnimation (s->texinfo->texture);
					GL_Bind (t->gl_texturenum);
					DrawGLPoly(t, s->lightmaptexturenum, s->polys, dlight, NULL, false);
				}
			}
		}
	}

	//draw entities
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = &cl_visedicts[i];

		if (currententity->model->type == mod_alias ||
			currententity->model->type == mod_brush ||
			currententity->model->type == mod_rdm)
		{
			if (currententity->model->type == mod_alias)
			{
				if (GL_AliasModelInLight(currententity, dlight))
				{
					GL_SetupCurEntityForDLight(dlight);
					R_DrawAliasModel (currententity, lightNum);
				}
			}
			else if (currententity->model->type == mod_rdm)
			{
				if (GL_RDMModelInLight(currententity, dlight))
				{
					GL_SetupCurEntityForDLight(dlight);
					Model_DrawRDMModel(currententity, lightNum);
				}
			}
			else
			{
				if (GL_BrushModelInLight(currententity, dlight))
				{
					GL_SetupCurEntityForDLight(dlight);
					R_DrawBrushModel (currententity, lightNum);
				}
			}
		}
	}

	//and, finally, the view model
	R_DrawViewModel (lightNum);

	//restore state
	g_glExt.glActiveTextureARB(GL_TEXTURE2_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE2_ARB);

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
	glDisable(GL_TEXTURE_GEN_Q);
	glMatrixMode(GL_MODELVIEW);

	glDisable(GL_TEXTURE_3D);

	g_glExt.glActiveTextureARB(GL_TEXTURE0_ARB);
	g_glExt.glClientActiveTextureARB(GL_TEXTURE0_ARB);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
	glDisable(GL_FRAGMENT_PROGRAM_ARB);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	//disable stencil/scissor
	glDisable(GL_STENCIL_TEST);
	glClear(GL_STENCIL_BUFFER_BIT);
	glDisable(GL_SCISSOR_TEST);

#if 0 //scissor test
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, SCIS_W, SCIS_H, 0, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glDisable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);
	glDepthMask(GL_FALSE);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	glColor4f(0.0f, 0.1f, 0.0f, 1.0f);
	glBegin(GL_QUADS);
		glVertex2f(dlight->frameScissor[0], dlight->frameScissor[1]);
		glVertex2f(dlight->frameScissor[0], dlight->frameScissor[1]+dlight->frameScissor[3]);
		glVertex2f(dlight->frameScissor[0]+dlight->frameScissor[2], dlight->frameScissor[1]+dlight->frameScissor[3]);
		glVertex2f(dlight->frameScissor[0]+dlight->frameScissor[2], dlight->frameScissor[1]);
	glEnd();
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);
	glEnable(GL_CULL_FACE);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
#endif

	return true;
}

/*
================
DrawTextureChains
================
*/
unsigned int g_chaincount = 0; //rww
void DrawTextureChains (void)
{
	int		i;
	msurface_t	*s;
	texture_t	*t;
	g_chaincount++;
	if (g_chaincount == 0)
	{ //overflow
		g_chaincount = 1;
	}

	if (!gl_texsort.value) {
		GL_DisableMultitexture();

		if (skychain) {
			R_DrawSkyChain(skychain);
			skychain = NULL;
		}

		return;
	} 
	else if (g_hasSun && skytexturenum >= 0)
	{ //rww - draw only depth for sky chains
		t = cl.worldmodel->textures[skytexturenum];
		if (t)
		{
			s = t->texturechain;
			if (s)
			{
				R_DrawSkyChain(s);
				if (!g_canUseTexFP || !gl_normalMaps.value)
				{
					t->texturechain = NULL;
				}
			}
		}
	}

	for (i=0 ; i<cl.worldmodel->numtextures ; i++)
	{
		t = cl.worldmodel->textures[i];
		if (!t)
			continue;
		s = t->texturechain;
		if (!s)
			continue;
		if (i == skytexturenum)
		{
			if (!g_hasSun)
			{
				R_DrawSkyChain (s);
				if (!g_canUseTexFP || !gl_normalMaps.value)
				{
					t->texturechain = NULL;
				}
			}
		}
		else if (i == mirrortexturenum && r_mirroralpha.value != 1.0)
		{
			R_MirrorChain (s);
			continue;
		}
		else
		{
			if ((s->flags & SURF_DRAWTURB) && r_wateralpha.value != 1.0)
				continue;	// draw translucent water later
			for ( ; s ; s=s->texturechain)
			{
				if (s->chaincount == g_chaincount)
				{
					break;
				}
				R_RenderBrushPoly (NULL, s, -1);
				s->chaincount = g_chaincount;
			}
			//rww - now clearing chains later
		}
	}
}

/*
=================
R_DrawBrushModel
=================
*/
void R_DrawBrushModel (entity_t *e, int lightNum) //rww - light
{
	int			i;
	int			k;
	msurface_t	*psurf;
	float		dot;
	mplane_t	*pplane;
	model_t		*clmodel;
	qboolean	rotated;

	currententity = e;
	currenttexture = -1;

	clmodel = e->model;

	if (e->cullFrame != r_framecount)
	{ //rww
		vec3_t		mins, maxs;
		if (e->angles[0] || e->angles[1] || e->angles[2])
		{
			rotated = true;
			for (i=0 ; i<3 ; i++)
			{
				mins[i] = e->origin[i] - clmodel->radius;
				maxs[i] = e->origin[i] + clmodel->radius;
			}
		}
		else
		{
			rotated = false;
			VectorAdd (e->origin, clmodel->mins, mins);
			VectorAdd (e->origin, clmodel->maxs, maxs);
		}

		e->culled = R_CullBox(mins, maxs);
		e->cullFrame = r_framecount;
	}

	if (e->culled)
	{ //rww - since these things go through passes for so many things, only do a cull check once per frame
		return;
	}

	if (lightNum < 0)
	{ //rww
		glColor3f (1,1,1);
		memset (lightmap_polys, 0, sizeof(lightmap_polys));
	}

	VectorSubtract (r_refdef.vieworg, e->origin, modelorg);
	if (rotated)
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (modelorg, temp);
		AngleVectors (e->angles, forward, right, up);
		modelorg[0] = DotProduct (temp, forward);
		modelorg[1] = -DotProduct (temp, right);
		modelorg[2] = DotProduct (temp, up);
	}

	psurf = &clmodel->surfaces[clmodel->firstmodelsurface];

// calculate dynamic lighting for bmodel if it's not an
// instanced model
	if (lightNum < 0)
	{ //rww
		if (clmodel->firstmodelsurface != 0 && !gl_flashblend.value)
		{
			for (k=0 ; k<MAX_DLIGHTS ; k++)
			{
				int low, high;
				if ((cl_dlights[k].die < cl.time) ||
					(!cl_dlights[k].radius))
					continue;

				GL_DlightHiLo(k, &low, &high);

				R_MarkLights (&cl_dlights[k], low, high, 
					clmodel->nodes + clmodel->hulls[0].firstclipnode);
			}
		}
	}

    glPushMatrix ();
e->angles[0] = -e->angles[0];	// stupid quake bug
	R_RotateForEntity (e);
e->angles[0] = -e->angles[0];	// stupid quake bug

	//
	// draw texture
	//
	for (i=0 ; i<clmodel->nummodelsurfaces ; i++, psurf++)
	{
	// find which side of the node we are on
		pplane = psurf->plane;

		dot = DotProduct (modelorg, pplane->normal) - pplane->dist;

	// draw the polygon
		if (((psurf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) ||
			(!(psurf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))
		{
			if (gl_texsort.value)
				R_RenderBrushPoly (e, psurf, lightNum);
			else
				R_DrawSequentialPoly (psurf);
		}
	}

	if (lightNum < 0)
	{ //rww
		R_BlendLightmaps ();
		memset (lightmap_polys, 0, sizeof(lightmap_polys));
		//rww - must clear after due to re-ordering of things, don't want these polys being
		//drawn again
	}

	glPopMatrix ();
}

/*
=============================================================

	WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node)
{
	int			c, side;
	mplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	float		planeDist; //rww
	double		dot;

	if (node->contents == CONTENTS_SOLID)
		return;		// solid

	if (node->visframe != r_visframecount)
		return;
	if (R_CullBox (node->minmaxs, node->minmaxs+3))
		return;
	
// if a leaf node, draw stuff
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *)node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

	// deal with model fragments in this leaf
		if (pleaf->efrags)
			R_StoreEfrags (&pleaf->efrags);

		return;
	}

// node is just a decision point, so go down the apropriate sides

// find which side of the node we are on
	plane = node->plane;

	planeDist = plane->dist; //rww
	if (node->numsurfaces)
	{ //rww
		c = node->numsurfaces;
		surf = cl.worldmodel->surfaces + node->firstsurface;
		while (c)
		{
			if (surf->numSurfSprites)
			{ //has surface sprites, extend the plane (horrible hack)
				planeDist -= 512.0f;
				break;
			}
			c--;
			surf++;
		}
	}

	switch (plane->type)
	{
	case PLANE_X:
		dot = modelorg[0] - planeDist; //rww
		break;
	case PLANE_Y:
		dot = modelorg[1] - planeDist; //rww
		break;
	case PLANE_Z:
		dot = modelorg[2] - planeDist; //rww
		break;
	default:
		dot = DotProduct (modelorg, plane->normal) - planeDist; //rww
		break;
	}

	if (dot >= 0)
		side = 0;
	else
		side = 1;

// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

// draw stuff
	c = node->numsurfaces;

	if (c)
	{
		surf = cl.worldmodel->surfaces + node->firstsurface;

		if (dot < 0 -BACKFACE_EPSILON)
			side = SURF_PLANEBACK;
		else if (dot > BACKFACE_EPSILON)
			side = 0;
		{
			for ( ; c ; c--, surf++)
			{
				if (surf->visframe != r_framecount)
					continue;

				// don't backface underwater surfaces, because they warp
//				if ( !(surf->flags & SURF_UNDERWATER) && ( (dot < 0) ^ !!(surf->flags & SURF_PLANEBACK)) )
//					continue;		// wrong side
				if ( !(((r_viewleaf->contents==CONTENTS_EMPTY && (surf->flags & SURF_UNDERWATER)) ||
					(r_viewleaf->contents!=CONTENTS_EMPTY && !(surf->flags & SURF_UNDERWATER)))
					&& !(surf->flags & SURF_DONTWARP)) && ( (dot < 0) ^ !!(surf->flags & SURF_PLANEBACK)) )
					continue;		// wrong side

				// if sorting by texture, just store it out
				if (gl_texsort.value)
				{
					if (!mirror
					|| surf->texinfo->texture != cl.worldmodel->textures[mirrortexturenum])
					{
						surf->texturechain = surf->texinfo->texture->texturechain;
						surf->texinfo->texture->texturechain = surf;
					}
				} else if (surf->flags & SURF_DRAWSKY) {
					surf->texturechain = skychain;
					skychain = surf;
				} else if (surf->flags & SURF_DRAWTURB) {
					surf->texturechain = waterchain;
					waterchain = surf;
				} else
					R_DrawSequentialPoly (surf);

			}
		}

	}

// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}



/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (void)
{
	entity_t	ent;

	memset (&ent, 0, sizeof(ent));
	ent.model = cl.worldmodel;

	VectorCopy (r_refdef.vieworg, modelorg);

	currententity = &ent;
	currenttexture = -1;

	glColor3f (1,1,1);
	memset (lightmap_polys, 0, sizeof(lightmap_polys));
#ifdef QUAKE2
	R_ClearSkyBox ();
#endif

	R_RecursiveWorldNode (cl.worldmodel->nodes);


	if (g_hasSun && g_canUseOccQuery && gl_bloom.value &&
		g_canUseTexRect && g_canUseFBO)
	{ //rww
		glDisable(GL_FOG);
		g_chaincount = GL_DrawSceneDepth(g_chaincount, false, false);
		R_DrawSunSky(true);
		glClear(GL_DEPTH_BUFFER_BIT);
		glEnable(GL_FOG);
	}
	else
	{
		glDisable(GL_FOG);
		R_DrawSunSky(true);
		glEnable(GL_FOG);
	}

		DrawTextureChains ();

	R_BlendLightmaps ();

	R_PushDlights (); //rww - push dlights after determining vis

#ifdef QUAKE2
	R_DrawSkyBox ();
#endif
}


/*
===============
R_MarkLeaves
===============
*/
void R_MarkLeaves (void)
{
	byte	*vis;
	mnode_t	*node;
	int		i;
	byte	solid[4096];

	if (r_oldviewleaf == r_viewleaf && !r_novis.value)
		return;
	
	if (mirror)
		return;

	r_visframecount++;
	r_oldviewleaf = r_viewleaf;

	if (r_novis.value)
	{
		vis = solid;
		memset (solid, 0xff, (cl.worldmodel->numleafs+7)>>3);
	}
	else
		vis = Mod_LeafPVS (r_viewleaf, cl.worldmodel);
		
	for (i=0 ; i<cl.worldmodel->numleafs ; i++)
	{
		if (vis[i>>3] & (1<<(i&7)))
		{
			node = (mnode_t *)&cl.worldmodel->leafs[i+1];
			do
			{
				if (node->visframe == r_visframecount)
					break;
				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}
}



/*
=============================================================================

  LIGHTMAP ALLOCATION

=============================================================================
*/

// returns a texture number and the position inside it
int AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	int		texnum;

	for (texnum=0 ; texnum<MAX_LIGHTMAPS ; texnum++)
	{
		best = BLOCK_HEIGHT;

		for (i=0 ; i<BLOCK_WIDTH-w ; i++)
		{
			best2 = 0;

			for (j=0 ; j<w ; j++)
			{
				if (allocated[texnum][i+j] >= best)
					break;
				if (allocated[texnum][i+j] > best2)
					best2 = allocated[texnum][i+j];
			}
			if (j == w)
			{	// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > BLOCK_HEIGHT)
			continue;

		for (i=0 ; i<w ; i++)
			allocated[texnum][*x + i] = best + h;

		return texnum;
	}

	Sys_Error ("AllocBlock: full");
	return 0;
}


mvertex_t	*r_pcurrentvertbase;
model_t		*currentmodel;

int	nColinElim;

//rww
glpoly_t *GL_AllocPoly(int numVerts)
{
	const int numTris = numVerts-2;
	const int polySize = sizeof(glpoly_t) + ((numVerts-4) * VERTEXSIZE*sizeof(float))/*verts*/ +
		(sizeof(float)*3*numVerts)/*vertNormals*/ + (sizeof(modelTan_t)*numVerts)/*tangents*/ +
		(sizeof(int)*3*numTris)/*tris*/ + (sizeof(float)*3*numVerts)/*batchVerts*/ +
		(sizeof(float)*2*numVerts)/*batchTC*/ + (sizeof(float)*2*numVerts)/*batchTCLM*/ +
		(sizeof(float)*3*numVerts)/*lightVectors*/ + (sizeof(float)*3*numVerts)/*haVectors*/ +
		(sizeof(modelTriNeighbor_t)*numTris);
	glpoly_t *p = Hunk_Alloc(polySize);
	p->numTris = numTris;
	p->tris = (int *)(((byte *)p)+sizeof(glpoly_t) + ((numVerts-4) * VERTEXSIZE*sizeof(float)));
	p->vertNormals = (float *)(((byte *)p->tris)+(sizeof(int)*3*numTris));
	p->tangents = (modelTan_t *)(((byte *)p->vertNormals)+(sizeof(float)*3*numVerts));
	p->batchVerts = (float *)(((byte *)p->tangents)+(sizeof(modelTan_t)*numVerts));
	p->batchTC = (float *)(((byte *)p->batchVerts)+(sizeof(float)*3*numVerts));
	p->batchTCLM = (float *)(((byte *)p->batchTC)+(sizeof(float)*2*numVerts));
	p->lightVectors = (float *)(((byte *)p->batchTCLM)+(sizeof(float)*2*numVerts));
	p->haVectors = (float *)(((byte *)p->lightVectors)+(sizeof(float)*3*numVerts));
	p->neighbors = (modelTriNeighbor_t *)(((byte *)p->haVectors)+(sizeof(float)*3*numVerts));

	return p;
}

//rww
static __inline GL_ClosestPointOnLine(float *a, float *b, float *p, float *out)
{
	float c[3];
	float v[3];
	float d, t;

	VectorSubtract(p, a, c);
	VectorSubtract(b, a, v);
	d = Length(v);
	FastVecNorm(v);
	t = DotProduct(v, c);
	if (t < 0.0f)
	{
		VectorCopy(a, out);
	}
	if (t > d)
	{
		VectorCopy(b, out);
	}
	FastVecScale(v, t);
	VectorAdd(a, v, out);
}

//rww
static __inline void GL_Orthogonalize(float *v1, float *v2, float *res)
{
	float v[3];
	float vx[3];
	VectorCopy(v1, vx);
	FastVecScale(vx, -1.0f);
	GL_ClosestPointOnLine(v1, vx, v2, v);
	VectorSubtract(v2, v, res);
	FastVecNorm(res);
}

//rww
void GL_CreateTangentMatrices(int numVerts, modelVert_t *verts, int numFaces, modelTriFace_t *faces,
											  modelTan_t *argTangentData, modelTexCoord_t *coords, float *vertNormals)
{
	int i = 0;
	modelTriFace_t *face = faces;
	modelTexCoord_t *coordBase = coords;

	modelTan_t *tan = argTangentData;

	memset(tan, 0, sizeof(modelTan_t)*numVerts);

	while (i < numFaces)
	{
		modelVert_t *v1, *v2, *v3;
		modelTexCoord_t *t1, *t2, *t3;
		modelTan_t *tan1, *tan2, *tan3;
		float *v1use, *v2use, *v3use;
		float planes[3][3];
		int x;

		v1 = (verts+face->a);
		v2 = (verts+face->b);
		v3 = (verts+face->c);

		tan1 = (tan+face->a);
		tan2 = (tan+face->b);
		tan3 = (tan+face->c);

		v1use = &v1->x;
		v2use = &v2->x;
		v3use = &v3->x;

		t1 = (coordBase+face->a);
		t2 = (coordBase+face->b);
		t3 = (coordBase+face->c);

		x = 0;
		while (x < 3)
		{
			float sc1[3], sc2[3];
			sc1[0] = v2use[x]-v1use[x];
			sc1[1] = t2->u-t1->u;
			sc1[2] = t2->v-t1->v;

			sc2[0] = v3use[x]-v1use[x];
			sc2[1] = t3->u-t1->u;
			sc2[2] = t3->v-t1->v;

			CrossProduct(sc1, sc2, planes[x]);
			x++;
		}

		if (planes[0][0] != 0.0f &&
			planes[1][0] != 0.0f &&
			planes[2][0] != 0.0f)
		{ //otherwise would divide by 0, so, looks like we have an invalid tangent.
			float dsdt[3];
			float dtdt[3];

			dsdt[0] = -planes[0][1]/planes[0][0];
			dsdt[1] = -planes[1][1]/planes[1][0];
			dsdt[2] = -planes[2][1]/planes[2][0];
			dtdt[0] = -planes[0][2]/planes[0][0];
			dtdt[1] = -planes[1][2]/planes[1][0];
			dtdt[2] = -planes[2][2]/planes[2][0];

			VectorAdd(tan1->tangentMatrix.x1, dsdt, tan1->tangentMatrix.x1);
			VectorAdd(tan1->tangentMatrix.x2, dtdt, tan1->tangentMatrix.x2);

			VectorAdd(tan2->tangentMatrix.x1, dsdt, tan2->tangentMatrix.x1);
			VectorAdd(tan2->tangentMatrix.x2, dtdt, tan2->tangentMatrix.x2);

			VectorAdd(tan3->tangentMatrix.x1, dsdt, tan3->tangentMatrix.x1);
			VectorAdd(tan3->tangentMatrix.x2, dtdt, tan3->tangentMatrix.x2);
		}

		tan1->numNormals++;
		tan2->numNormals++;
		tan3->numNormals++;

		face++;
		i++;
	}

	i = 0;
	while (i < numVerts)
	{
		float *vNorm = vertNormals+(i*3);

		if (tan->numNormals)
		{
			tan->tangentMatrix.x1[0] /= tan->numNormals;
			tan->tangentMatrix.x1[1] /= tan->numNormals;
			tan->tangentMatrix.x1[2] /= tan->numNormals;

			tan->tangentMatrix.x2[0] /= tan->numNormals;
			tan->tangentMatrix.x2[1] /= tan->numNormals;
			tan->tangentMatrix.x2[2] /= tan->numNormals;

			VectorCopy(vNorm, tan->tangentMatrix.x3);
			GL_Orthogonalize(tan->tangentMatrix.x3, tan->tangentMatrix.x1, tan->tangentMatrix.x1);
			GL_Orthogonalize(tan->tangentMatrix.x3, tan->tangentMatrix.x2, tan->tangentMatrix.x2);
		}
		else
		{
			tan->tangentMatrix.x1[0] = 1.0f;
			tan->tangentMatrix.x1[1] = 0.0f;
			tan->tangentMatrix.x1[2] = 0.0f;

			tan->tangentMatrix.x2[0] = 0.0f;
			tan->tangentMatrix.x2[1] = 1.0f;
			tan->tangentMatrix.x2[2] = 0.0f;

			tan->tangentMatrix.x3[0] = 0.0f;
			tan->tangentMatrix.x3[1] = 0.0f;
			tan->tangentMatrix.x3[2] = 1.0f;
		}

		tan++;
		i++;
	}
}

//rww
void GL_GenerateVertexNormals(float *normals, int numVerts, float *verts, int numTris, int *tris)
{
	int i;
	int j;
	float *vert = verts;
	int *tri;
	float *normal = normals;
	float *triPlanes = (float *)_alloca(sizeof(float)*4*numTris);

	//calculate triangle planes
	for (i = 0; i < numTris; i++)
	{
		tri = tris+(i*3);
		PlaneEquation(verts+(tri[0]*3), verts+(tri[1]*3), verts+(tri[2]*3), triPlanes+(i*4));
	}

	i = 0;
	while (i < numVerts)
	{
		normal[0] = 0.0f;
		normal[1] = 0.0f;
		normal[2] = 0.0f;
		j = 0;
		tri = tris;
		while (j < numTris)
		{
			float *v1 = verts+(tri[0]*3);
			float *v2 = verts+(tri[1]*3);
			float *v3 = verts+(tri[2]*3);

			if (vert == v1 ||
				vert == v2 ||
				vert == v3)
			{ //this vert is attached to this face, so add its plane into the will-be average
				float *triPlane = triPlanes+(j*4);
				VectorAdd(normal, triPlane, normal);
			}

			tri += 3;
			j++;
		}
		FastVecNorm(normal);

		normal += 3;
		vert += 3;
		i++;
	}
}

//rww - calculates light and potentially halfangle vectors for a glpoly
//int RecursiveLightPoint (mnode_t *node, vec3_t start, vec3_t end);
void GL_CalcPolyLightVectors(glpoly_t *poly, dlight_t *dlight, entity_t *e, qboolean getHalfAngle)
{
	int i;
	float *vertLightPos = NULL;

	if (dlight)
	{
		if (e)
		{
			vertLightPos = dlight->localPos;
		}
		else
		{
			vertLightPos = dlight->origin;
		}
	}
	else if (g_hasSun)
	{
		if (e)
		{
			GL_SetupCurEntityForDLight(NULL);
			vertLightPos = currententity->sunLocal;
		}
		else
		{
			vertLightPos = g_sunPos;
		}
	}

	for (i = 0; i < poly->numverts; i++)
	{
		float localLightVec[3];
		modelTan_t *tan = poly->tangents+i;
		float *vert = poly->batchVerts+(i*3);
		float *lightVec = poly->lightVectors+(i*3);
		//interesting, but not computationally practical and buggy
		/*
		float *normal = poly->vertNormals+(i*3);
		int j;
		int bestLightIndex = -1;
		int bestLightVal = -1;
		const float lgSize = 64.0f;
		const float vertLightOffs[8][3] =
		{
			{-lgSize,	-lgSize,	-lgSize		},
			{-lgSize,	lgSize,		-lgSize		},
			{lgSize,	-lgSize,	-lgSize		},
			{lgSize,	lgSize,		-lgSize		},
			{-lgSize,	-lgSize,	lgSize		},
			{-lgSize,	lgSize,		lgSize		},
			{lgSize,	-lgSize,	lgSize		},
			{lgSize,	lgSize,		lgSize		},
		};
		for (j = 0; j < 8; j++)
		{
			int l;
			float d;
			float v[3], vd[3];

			VectorAdd(vert, vertLightOffs[j], v);
			d = DotProduct(vertLightOffs[j], normal);
			l = RecursiveLightPoint(cl.worldmodel->nodes, vert, v);
			if (d > 0.0f && l > bestLightVal)
			{
				bestLightIndex = j;
				bestLightVal = l;
			}
		}
		if (bestLightIndex == -1)
		{
			VectorScale(normal, 8.0f, vertLightPos);
			VectorAdd(vert, vertLightPos, vertLightPos);
		}
		else
		{
			VectorAdd(vert, vertLightOffs[bestLightIndex], vertLightPos);
		}
		*/

		if (vertLightPos)
		{
			VectorSubtract(vertLightPos, vert, localLightVec);
			lightVec[0] = DotProduct(localLightVec, tan->tangentMatrix.x1);
			lightVec[1] = DotProduct(localLightVec, tan->tangentMatrix.x2);
			lightVec[2] = -DotProduct(localLightVec, tan->tangentMatrix.x3);
			if (getHalfAngle)
			{
				float *haVec = poly->haVectors+(i*3);
				float haPreTrans[3];

				FastVecNorm(localLightVec);
				if (e)
				{
					VectorSubtract(e->viewLocal, vert, haPreTrans);
				}
				else
				{
					VectorSubtract(r_refdef.vieworg, vert, haPreTrans);
				}
				FastVecNorm(haPreTrans);
				VectorAdd(haPreTrans, localLightVec, haPreTrans);
				haVec[0] = DotProduct(haPreTrans, tan->tangentMatrix.x1);
				haVec[1] = DotProduct(haPreTrans, tan->tangentMatrix.x2);
				haVec[2] = -DotProduct(haPreTrans, tan->tangentMatrix.x3);
			}
		}
		else
		{ //dumb. do something clever.
			lightVec[0] = 0.0f;
			lightVec[1] = 0.0f;
			lightVec[2] = 1.0f;
			if (getHalfAngle)
			{
				float *haVec = poly->haVectors+(i*3);

				haVec[0] = 0.0f;
				haVec[1] = 0.0f;
				haVec[2] = 1.0f;
			}
		}

		//todo - using local viewpos, set halfangles
	}
}

//rww
void GL_CalculateMeshNeighbors(int numTris, int *tris, modelTriNeighbor_t *neighbors)
{
	int m;
	for (m = 0; m < numTris; m++)
	{
		int *tri = tris+(m*3);
		int *neighbor = (int *)(neighbors+m);

		int *n;
		int i = 0;
		int numN = 0;
		int j, k;
		int vertA1, vertA2, vertB1, vertB2;

		neighbor[0] = neighbor[1] = neighbor[2] = -1;

		while (i < numTris)
		{
			n = (int *)tris+(i*3);

			if (n != tri)
			{
				j = 0;
				while (j < 3)
				{
					vertA1 = tri[j];
					vertA2 = tri[(j+1)%3];

					k = 0;
					while (k < 3)
					{
						vertB1 = n[k];
						vertB2 = n[(k+1)%3];

						if ((vertA1 == vertB1 && vertA2 == vertB2) || (vertA1 == vertB2 && vertA2 == vertB1))
						{
							if (neighbor[j] == -1)
							{
								neighbor[j] = i;
							}
							else
							{ //this edge is shared by more than 2 tris apparently.
							  //this seems to actually happen at times, which is why
							  //we are bothering to check for it.
								neighbor[j] = -2;
							}
						}

						k++;
					}

					j++;
				}
			}

			i++;
		}

		j = 0;
		while (j < 3)
		{ //mark any -2 as -1 now since we're done
			if (neighbor[j] == -2)
			{
				neighbor[j] = -1;
			}
			j++;
		}

	}
}

//rww
void GL_VertMinsMaxs(int numVerts, float *verts, float *mins, float *maxs)
{
	int i, j;
	mins[0] = mins[1] = mins[2] = -16.0f; //hack
	maxs[0] = maxs[1] = maxs[2] = 16.0f; //hack
	for (i = 0; i < numVerts; i++)
	{
		float *v = verts+(i*3);
		for (j = 0; j < 3; j++)
		{
			if (v[j] < mins[j])
			{
				mins[j] = v[j];
			}
			if (v[j] > maxs[j])
			{
				maxs[j] = v[j];
			}
		}
	}
}

//rww - calculate normals and such for a glpoly
float RadiusFromBounds (vec3_t mins, vec3_t maxs);
void GL_CalcPolyEx(glpoly_t *poly)
{
	int *tri = poly->tris;
	float lmins[3], lmaxs[3];
	float mins[3], maxs[3], center[3];
	int i, j;

	//create vert/tc batches
	for (i = 0; i < poly->numverts; i++)
	{
		float *v = poly->verts[i];
		float *vert = poly->batchVerts+(i*3);
		float *tc = poly->batchTC+(i*2);
		float *tclm = poly->batchTCLM+(i*2);
		vert[0] = v[0];
		vert[1] = v[1];
		vert[2] = v[2];
		tc[0] = v[3];
		tc[1] = v[4];
		tclm[0] = v[5];
		tclm[1] = v[6];
	}

	//create triangles for batch rendering
	tri[0] = 0;
	tri[1] = 1;
	tri[2] = 2;
	tri += 3;
	j = 3;
	for (i = 1; i < poly->numTris; i++, tri += 3)
	{
		tri[0] = j;
		tri[1] = 0;
		tri[2] = j-1;
		j++;
	}

	//calc vertex normals
	GL_GenerateVertexNormals(poly->vertNormals, poly->numverts, poly->batchVerts, poly->numTris, poly->tris);

	//generate tangents
	GL_CreateTangentMatrices(poly->numverts, (modelVert_t *)poly->batchVerts, poly->numTris, (modelTriFace_t *)poly->tris, poly->tangents,
		(modelTexCoord_t *)poly->batchTC, poly->vertNormals);

	//create neighbors
	GL_CalculateMeshNeighbors(poly->numTris, poly->tris, poly->neighbors);

	//get radius
	GL_VertMinsMaxs(poly->numverts, poly->batchVerts, mins, maxs);
	center[0] = mins[0] + (maxs[0]-mins[0])/2.0f;
	center[1] = mins[1] + (maxs[1]-mins[1])/2.0f;
	center[2] = mins[2] + (maxs[2]-mins[2])/2.0f;
	VectorSubtract(mins, center, lmins);
	VectorSubtract(maxs, center, lmaxs);
	poly->radius = RadiusFromBounds(lmins, lmaxs);
}

//rww
qboolean GL_FlatFace(float *v1, float *v2, float *v3)
{
	int i, j;
	for (i = 0; i < 3; i++)
	{
		if (v1[i] == v2[i] && v2[i] == v3[i])
		{
			for (j = 0; j < 3; j++)
			{
				if (j == i)
				{
					continue;
				}
				if (v1[j] == v2[j] && v2[j] == v3[j])
				{ //flat on another axis
					return true;
				}
			}
			//if the j loop didn't hit, this isn't isn't a flat face
			return false;
		}
	}
	return false;
}

//rww - adapted from http://www.worldserver.com/turk/opensource/InvSqrt.c.txt
/* Copyright (C) 1986-1999 Ken Turkowski. <turk@computer.org>
 *
 * All rights reserved.
 *
 * Warranty Information
 *  Even though I have reviewed this software, I makes no warranty
 *  or representation, either express or implied, with respect to this
 *  software, its quality, accuracy, merchantability, or fitness for a
 *  particular purpose.  As a result, this software is provided "as is,"
 *  and you, its user, are assuming the entire risk as to its quality
 *  and accuracy.
 *
 * This code may be used and freely distributed as long as it includes
 *	this copyright notice and the above warranty information.
 *
 ********************************************************************************
 * InvSqrt calculates the inverse sqaure root of its argument.
 ********************************************************************************/
/* Specified parameters */
#define LOOKUP_BITS    6   /* Number of mantissa bits for lookup */
#define EXP_POS       23   /* Position of the exponent */
#define EXP_BIAS     127   /* Bias of exponent */
/* The mantissa is assumed to be just down from the exponent */

/* Derived parameters */
#define LOOKUP_POS   (EXP_POS-LOOKUP_BITS)  /* Position of mantissa lookup */
#define SEED_POS     (EXP_POS-8)            /* Position of mantissa seed */
#define TABLE_SIZE   (2 << LOOKUP_BITS)     /* Number of entries in table */
#define LOOKUP_MASK  (TABLE_SIZE - 1)           /* Mask for table input */
#define GET_EXP(a)   (((a) >> EXP_POS) & 0xFF)  /* Extract exponent */
#define SET_EXP(a)   ((a) << EXP_POS)           /* Set exponent */
#define GET_EMANT(a) (((a) >> LOOKUP_POS) & LOOKUP_MASK)  /* Extended mantissa
                                                           * MSB's */
#define SET_MANTSEED(a) (((unsigned long)(a)) << SEED_POS)  /* Set mantissa
                                                             * 8 MSB's */

static unsigned char iSqrt[TABLE_SIZE];
static qboolean madeSqrtTable = false;

union _flint {
    unsigned long    i;
    float            f;
} _fi, _fo;

static void GL_MakeInverseSqrtLookupTable(void)
{
    register long f;
    register unsigned char *h;
    union _flint fi, fo;

    h = iSqrt;
    for (f = 0, h = iSqrt; f < TABLE_SIZE; f++) {
        fi.i = ((EXP_BIAS-1) << EXP_POS) | (f << LOOKUP_POS);
        fo.f = 1.0 / sqrt(fi.f);
        *h++ = ((fo.i + (1<<(SEED_POS-2))) >> SEED_POS) & 0xFF; /* rounding */
    }
    iSqrt[TABLE_SIZE / 2] = 0xFF;    /* Special case for 1.0 */
	madeSqrtTable = true;
}
float GL_InvSqrt(float x)
{
    unsigned long a = ((union _flint*)(&x))->i;
    float arg = x;
    union _flint seed;
    float r;

    if (!madeSqrtTable) GL_MakeInverseSqrtLookupTable();

    seed.i = SET_EXP(((3*EXP_BIAS-1) - GET_EXP(a)) >> 1)
           | SET_MANTSEED(iSqrt[GET_EMANT(a)]);

    /* Seed: accurate to LOOKUP_BITS */
    r = seed.f;

    /* First iteration: accurate to 2*LOOKUP_BITS */
    r = (3.0 - r * r * arg) * r * 0.5;

    /* Second iteration: accurate to 4*LOOKUP_BITS */
    r = (3.0 - r * r * arg) * r * 0.5;

    return(r);
}

//rww
void GL_AxisForNormal(float *normal, float *fwd, float *right, float *up)
{
	float f = normal[0]*normal[0] + normal[1]*normal[1];
	float t[3];
	
	if (f == 0.0f)
	{
		t[0] = 1.0f;
		t[1] = 0.0f;
		t[2] = 0.0f;
	}
	else
	{
		f = GL_InvSqrt(f);
		t[0] = -normal[1]*f;
		t[1] = normal[0]*f;
		t[2] = 0.0f;
	}

	if (fwd)
	{
		VectorCopy(normal, fwd);
	}

	if (right)
	{
		VectorCopy(t, right);
	}

	if (up)
	{
		CrossProduct(normal, t, up);
	}
}

//rww
//todo move into math along with some of these other functions
void GL_RotationMatrix(float phi, int axis, float *mat)
{
	memset(mat, 0, sizeof(float)*16);

	switch (axis)
	{
	case 0: //x
		mat[0+0] = 1.0f;
		mat[0+1] = 0.0f;
		mat[0+2] = 0.0f;
		mat[4+0] = 0.0f;
		mat[4+1] = cos(phi);
		mat[4+2] = sin(phi);
		mat[8+0] = 0.0f;
		mat[8+1] = -sin(phi);
		mat[8+2] = cos(phi);
		break;
	case 1: //y
		mat[0+0] = cos(phi);
		mat[0+1] = 0.0f;
		mat[0+2] = sin(phi);
		mat[4+0] = 0.0f;
		mat[4+1] = 1.0f;
		mat[4+2] = 0.0f;
		mat[8+0] = -sin(phi);
		mat[8+1] = 0.0f;
		mat[8+2] = cos(phi);
		break;
	case 2: //z
		mat[0+0] = cos(phi);
		mat[0+1] = sin(phi);
		mat[0+2] = 0.0f;
		mat[4+0] = -sin(phi);
		mat[4+1] = cos(phi);
		mat[4+2] = 0.0f;
		mat[8+0] = 0.0f;
		mat[8+1] = 0.0f;
		mat[8+2] = 1.0f;
		break;

	default:
		break;
	}

	mat[15] = 1.0f;
}

//rww
void GL_ConfinePosToEdge(float *pos, float *v1, float *v2, float *plane, float radius)
{
	float v3[3];
	float edgePlane[4];
	float side;
	
	//the edge plane can be formed using a third position which is the first point
	//on the edge extruded upon the the face plane normal
	VectorAdd(v1, plane, v3);
	PlaneEquation(v1, v2, v3, edgePlane);
	PlaneNormalize(edgePlane);

	side = edgePlane[0]*pos[0] + edgePlane[1]*pos[1] + edgePlane[2]*pos[2] + edgePlane[3];
	if (side < radius)
	{ //move it back in by the radius
		side -= radius;
		pos[0] -= edgePlane[0]*side;
		pos[1] -= edgePlane[1]*side;
		pos[2] -= edgePlane[2]*side;
	}
}

//rww
void GL_ConfinePosToPolyEdges(glpoly_t *poly, float *pos, float *plane, float radius)
{
	int i, j, k, m;

	for (i = 0; i < poly->numTris; i++)
	{ //cycle through all tris to find edges
		int *tri = poly->tris+(i*3);
		float *v[3];
		qboolean edges[3];
		for (k = 0; k < 3; k++)
		{
			v[k] = poly->batchVerts+(tri[k]*3);
			edges[k] = true;
		}
		for (j = 0; j < poly->numTris; j++)
		{ //find external edges by comparing shared verts using other tris on the poly
			int *ntri;
			if (i == j)
			{
				continue;
			}
			ntri = poly->tris+(j*3);
			for (k = 0; k < 3; k++)
			{
				for (m = 0; m < 3; m++)
				{
					if (ntri[k] == tri[m])
					{ //shares one vert
						if (tri[(m+1)%3] == ntri[(k+1)%3] ||
							tri[(m+1)%3] == ntri[(k+2)%3])
						{ //shares the vert on the edge with the first vert
							edges[m] = false;
						}
					}
				}
			}
		}
		for (k = 0; k < 3; k++)
		{ //confine to the edges which were not internal
			if (edges[k])
			{
				GL_ConfinePosToEdge(pos, v[k], v[(k+1)%3], plane, radius);
			}
		}
	}
}

//rww
qboolean GL_CheckOutsidePoly(glpoly_t *poly, float *pos, float *plane, float radius)
{
	int i, j, k, m;

	for (i = 0; i < poly->numTris; i++)
	{ //cycle through all tris to find edges
		int *tri = poly->tris+(i*3);
		float *v[3];
		qboolean edges[3];
		for (k = 0; k < 3; k++)
		{
			v[k] = poly->batchVerts+(tri[k]*3);
			edges[k] = true;
		}
		for (j = 0; j < poly->numTris; j++)
		{ //find external edges by comparing shared verts using other tris on the poly
			int *ntri;
			if (i == j)
			{
				continue;
			}
			ntri = poly->tris+(j*3);
			for (k = 0; k < 3; k++)
			{
				for (m = 0; m < 3; m++)
				{
					if (ntri[k] == tri[m])
					{ //shares one vert
						if (tri[(m+1)%3] == ntri[(k+1)%3] ||
							tri[(m+1)%3] == ntri[(k+2)%3])
						{ //shares the vert on the edge with the first vert
							edges[m] = false;
						}
					}
				}
			}
		}
		for (k = 0; k < 3; k++)
		{ //check the edges which were not internal
			if (edges[k])
			{
				float v3[3];
				float edgePlane[4];
				float side;
				
				//the edge plane can be formed using a third position which is the first point
				//on the edge extruded upon the the face plane normal
				VectorAdd(v[k], plane, v3);
				PlaneEquation(v[k], v[(k+1)%3], v3, edgePlane);
				PlaneNormalize(edgePlane);

				side = edgePlane[0]*pos[0] + edgePlane[1]*pos[1] + edgePlane[2]*pos[2] + edgePlane[3];
				if (side < radius)
				{ //it's outside
					return true;
				}
			}
		}
	}

	return false;
}

//rww
extern cvar_t gl_surfSpriteDensity;
void GL_GenerateSurfSprites(msurface_t *fa)
{
	glpoly_t *poly;
	int i;
	float edgeLengths = 0.0f;
	float spriteDensity = gl_surfSpriteDensity.value;//16.0f;//32.0f;
	float rotMat[16], surfMat[16];
	float realPlane[4];
	int *tri;
	int sprTri = 0;
	int sprTriCount = 0;
	int *numPerTri;
	int numTotal = 0;
	float *v1;
	float *v2;
	float *v3;

	if (fa->numSurfSprites)
	{ //already has surface sprites
		return;
	}

	if (stricmp(fa->texinfo->texture->name, "d1_grass"))
	{ //does not use surface sprites
		return;
	}

	//evaluate the surface area to determine how many sprites we want
	poly = fa->polys;
	numPerTri = _alloca(sizeof(int)*poly->numTris);
	memset(numPerTri, 0, sizeof(int)*poly->numTris);
	for (i = 0; i < poly->numTris; i++)
	{
		float t[3];
		float l = 0.0f;
		int *n = numPerTri+i;
		float edgeLen;
		tri = poly->tris+(i*3);
		v1 = poly->batchVerts+(tri[0]*3);
		v2 = poly->batchVerts+(tri[1]*3);
		v3 = poly->batchVerts+(tri[2]*3);
		if (GL_FlatFace(v1, v2, v3))
		{
			continue;
		}

		VectorSubtract(v2, v1, t);
		edgeLen = Length(t);
		l = edgeLen;

		VectorSubtract(v3, v1, t);
		edgeLen = Length(t);
		l *= edgeLen;

		VectorSubtract(v2, v3, t);
		edgeLen = Length(t);
		l *= edgeLen;

		*n = (int)(sqrtf(l)*spriteDensity);
		numTotal += *n;
	}

	//use the averaged edge length to determine a number
	fa->numSurfSprites = numTotal;
	if (fa->numSurfSprites <= 0)
	{ //didn't generate any sprites
		return;
	}

	fa->surfSprites = (surfSprite_t *)Hunk_Alloc(sizeof(surfSprite_t)*fa->numSurfSprites + sizeof(float)*16);
	fa->surfSpriteMat = (float *)(fa->surfSprites+fa->numSurfSprites);

	if (poly->numTris > 1)
	{
		tri = poly->tris+(1*3);
	}
	else
	{
		tri = poly->tris;
	}
	v1 = poly->batchVerts+(tri[0]*3);
	v2 = poly->batchVerts+(tri[1]*3);
	v3 = poly->batchVerts+(tri[2]*3);
	i = 0;
	while (GL_FlatFace(v1, v2, v3) && i < poly->numTris)
	{ //try to avoid faces which are flat on 2+ axises
		tri = poly->tris+(i*3);
		v1 = poly->batchVerts+(tri[0]*3);
		v2 = poly->batchVerts+(tri[1]*3);
		v3 = poly->batchVerts+(tri[2]*3);
		i++;
	}
	PlaneEquation(v3, v2, v1, realPlane);
	VectorNormalize(realPlane);

	//create a rotation matrix, form a matrix from the surface orientation, and form a final
	//rotated surface-relative orientation matrix
	GL_RotationMatrix(-90.0f*0.0174532925f, 1, rotMat);
	memset(surfMat, 0, sizeof(surfMat));
	GL_AxisForNormal(realPlane, &surfMat[0], &surfMat[4], &surfMat[8]);
	surfMat[15] = 1.0f;
	MatrixMultiply4x4(surfMat, rotMat, fa->surfSpriteMat);

	for (i = 0; i < fa->numSurfSprites; i++)
	{
		surfSprite_t *spr = fa->surfSprites+i;
		int grassIndex = rand()%num_grass_textures;
		int grassTex = grass_textures[grassIndex];
		int grassShadowTex = grass_shadow_textures[grassIndex];
		int flatTries = 0;
		int triN = *(numPerTri+sprTri);
		int posAttempts = 0;
		float p[3];
		if (sprTriCount >= triN)
		{
			sprTriCount = 0;
			sprTri++;
			if (sprTri >= poly->numTris)
			{
				Con_Printf("WARNING: GL_GenerateSurfSprites went out of range.\n");
				break;
			}
			continue;
		}
		sprTriCount++;
		tri = poly->tris+(sprTri*3);
		v1 = poly->batchVerts+(tri[0]*3);
		v2 = poly->batchVerts+(tri[1]*3);
		v3 = poly->batchVerts+(tri[2]*3);

		//get a random point in the triangle and confine it to the poly edges
		R_RandomPointInTri(v1, v2, v3, spr->basePos);
		GL_ConfinePosToPolyEdges(fa->polys, spr->basePos, realPlane, 18.0f);

		//for hopeless conditions, try new positions
		while (GL_CheckOutsidePoly(fa->polys, spr->basePos, realPlane, 3.8f) && posAttempts < 32)
		{
			R_RandomPointInTri(v1, v2, v3, spr->basePos);
			GL_ConfinePosToPolyEdges(fa->polys, spr->basePos, realPlane, 4.0f);
			posAttempts++;
		}
		if (posAttempts == 32)
		{ //failed completely
			fa->numSurfSprites--;
			i--;
			continue;
		}

		spr->texture = grassTex;
		spr->shadowTexture = grassShadowTex;
		
		spr->baseW = (float)(28+rand()%8);
		spr->baseH = (float)(28+rand()%8);
		spr->offRot = (float)(rand()%360);

		//start off with no wind
		spr->windEffect[0] = 0.0f;
		spr->windEffect[1] = 0.0f;
		spr->windVelocity[0] = 0.0f;
		spr->windVelocity[1] = 0.0f;

		p[0] = spr->basePos[0];
		p[1] = spr->basePos[1];
		p[2] = spr->basePos[2] + 8.0f;
		spr->baseColor = 0.2 + (((float)R_LightPoint(p) / 255.0f) * 0.25f);
	}
}

/*
================
BuildSurfaceDisplayList
================
*/
void BuildSurfaceDisplayList (msurface_t *fa)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	int			vertpage;
	float		*vec;
	float		s, t;
	glpoly_t	*poly;

// reconstruct the polygon
	pedges = currentmodel->edges;
	lnumverts = fa->numedges;
	vertpage = 0;

	//
	// draw texture
	//
	poly = GL_AllocPoly(lnumverts);
	poly->next = fa->polys;
	poly->flags = fa->flags;
	fa->polys = poly;
	poly->numverts = lnumverts;

	for (i=0 ; i<lnumverts ; i++)
	{
		lindex = currentmodel->surfedges[fa->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = r_pcurrentvertbase[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = r_pcurrentvertbase[r_pedge->v[1]].position;
		}
		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s /= fa->texinfo->texture->width;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t /= fa->texinfo->texture->height;

		VectorCopy (vec, poly->verts[i]);
		poly->verts[i][3] = s;
		poly->verts[i][4] = t;

		//
		// lightmap texture coordinates
		//
		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		s -= fa->texturemins[0];
		s += fa->light_s*16;
		s += 8;
		s /= BLOCK_WIDTH*16; //fa->texinfo->texture->width;

		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];
		t -= fa->texturemins[1];
		t += fa->light_t*16;
		t += 8;
		t /= BLOCK_HEIGHT*16; //fa->texinfo->texture->height;

		poly->verts[i][5] = s;
		poly->verts[i][6] = t;
	}

	//
	// remove co-linear points - Ed
	//
	if (!gl_keeptjunctions.value && !(fa->flags & SURF_UNDERWATER) )
	{
		for (i = 0 ; i < lnumverts ; ++i)
		{
			vec3_t v1, v2;
			float *prev, *this, *next;

			prev = poly->verts[(i + lnumverts - 1) % lnumverts];
			this = poly->verts[i];
			next = poly->verts[(i + 1) % lnumverts];

			VectorSubtract( this, prev, v1 );
			VectorNormalize( v1 );
			VectorSubtract( next, prev, v2 );
			VectorNormalize( v2 );

			// skip co-linear points
			#define COLINEAR_EPSILON 0.001
			if ((fabs( v1[0] - v2[0] ) <= COLINEAR_EPSILON) &&
				(fabs( v1[1] - v2[1] ) <= COLINEAR_EPSILON) && 
				(fabs( v1[2] - v2[2] ) <= COLINEAR_EPSILON))
			{
				int j;
				for (j = i + 1; j < lnumverts; ++j)
				{
					int k;
					for (k = 0; k < VERTEXSIZE; ++k)
						poly->verts[j - 1][k] = poly->verts[j][k];
				}
				--lnumverts;
				++nColinElim;
				// retry next vertex next time, which is now current vertex
				--i;
			}
		}
	}
	poly->numverts = lnumverts;

	GL_CalcPolyEx(poly); //rww
	GL_GenerateSurfSprites(fa); //rww
}

/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (msurface_t *surf)
{
	int		smax, tmax;
	byte	*base;

	if (surf->flags & (SURF_DRAWSKY|SURF_DRAWTURB))
		return;

	smax = (surf->extents[0]>>4)+1;
	tmax = (surf->extents[1]>>4)+1;

	surf->lightmaptexturenum = AllocBlock (smax, tmax, &surf->light_s, &surf->light_t);
	base = lightmaps + surf->lightmaptexturenum*lightmap_bytes*BLOCK_WIDTH*BLOCK_HEIGHT;
	base += (surf->light_t * BLOCK_WIDTH + surf->light_s) * lightmap_bytes;
	R_BuildLightMap (surf, base, BLOCK_WIDTH*lightmap_bytes);
}


/*
==================
GL_BuildLightmaps

Builds the lightmap texture
with all the surfaces from all brush models
==================
*/
void GL_BuildLightmaps (void)
{
	int		i, j;
	model_t	*m;

	memset (allocated, 0, sizeof(allocated));

	r_framecount = 1;		// no dlightcache

	if (!normalmap_default && g_canUseTexFP)
	{ //rww
		unsigned char *p;
		int k;

		normalmap_default = texture_extension_number;
		texture_extension_number++;

		GL_Bind(normalmap_default);
		p = (unsigned char *)malloc(8*8*4);
		for (k = 0; k < (8*8); k++)
		{ //create an image with all normals pointing up
			p[(k*4)+0] = 128;
			p[(k*4)+1] = 128;
			p[(k*4)+2] = 255;
			p[(k*4)+3] = 255;
		}
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 8, 8, 0, GL_RGBA, GL_UNSIGNED_BYTE, p);
		free(p);
	}
	if (!num_grass_textures)
	{ //rww
		grass_textures[0] = texture_extension_number;
		texture_extension_number++;
		grass_shadow_textures[0] = texture_extension_number;
		texture_extension_number++;
		num_grass_textures++;
		GL_UploadPNG("qdix/grass/grass01.png", grass_textures[0], true);
		GL_UploadPNG("qdix/grass/grass01_shadow.png", grass_shadow_textures[0], true);

		grass_textures[1] = texture_extension_number;
		texture_extension_number++;
		grass_shadow_textures[1] = grass_shadow_textures[0];
		num_grass_textures++;
		GL_UploadPNG("qdix/grass/grass02.png", grass_textures[1], true);
	}
	if (!attenmap_default && g_canUseTexFP)
	{ //rww
		const int attnSliceSize = 128;
		const int internalFormat = GL_INTENSITY8;
		byte *attnSlice = (byte *)malloc(attnSliceSize*sizeof(byte));
		byte *attnTex = (byte *)malloc((attnSliceSize*attnSliceSize*attnSliceSize)*sizeof(byte));
		byte sX, sY, sZ;
		int i, j, k, l;

		attenmap_default = texture_extension_number;
		texture_extension_number++;

		for(i = 0; i < attnSliceSize; i++)
		{ //fill in a slice
			const float center = (attnSliceSize/2.0f)-0.5f;
			float d = ((float)i-center)/center;
			d *= d;
			if (d > 1.0f)
			{
				d = 1.0f;
			}
			else if (d < 0.0f)
			{
				d = 0.0f;
			}

			attnSlice[i] = (byte)(d*255.0f);
		}
		attnSlice[0] = 255;
		attnSlice[attnSliceSize-1] = 255;

		l = 0;
		//put the slice into the 3d attenuation map
		for (i = 0; i < attnSliceSize; i++)
		{
			sX = attnSlice[i];

			for (j = 0; j < attnSliceSize; j++)
			{
				sY = attnSlice[j];

				for (k = 0; k < attnSliceSize; k++)
				{
					int t;

					sZ = attnSlice[k];
					t = sX+sY+sZ;
					if (t > 255)
					{
						t = 255;
					}
					t = (255-t);
			
					attnTex[l] = t;
					l++;
				}
			}
		}		

		glBindTexture(GL_TEXTURE_3D, attenmap_default);
		g_glExt.glTexImage3DEXT(GL_TEXTURE_3D, 0, internalFormat, attnSliceSize, attnSliceSize, attnSliceSize,
								0, GL_LUMINANCE, GL_UNSIGNED_BYTE, attnTex);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

		free(attnSlice);
		free(attnTex);
	}

	if (!lightmap_textures)
	{
		lightmap_textures = texture_extension_number;
		texture_extension_number += MAX_LIGHTMAPS;
	}

	gl_lightmap_format = GL_LUMINANCE;
	if (COM_CheckParm ("-lm_1"))
		gl_lightmap_format = GL_LUMINANCE;
	if (COM_CheckParm ("-lm_a"))
		gl_lightmap_format = GL_ALPHA;
	if (COM_CheckParm ("-lm_i"))
		gl_lightmap_format = GL_INTENSITY;
	if (COM_CheckParm ("-lm_2"))
		gl_lightmap_format = GL_RGBA4;
	if (COM_CheckParm ("-lm_4"))
		gl_lightmap_format = GL_RGBA;

	switch (gl_lightmap_format)
	{
	case GL_RGBA:
		lightmap_bytes = 4;
		break;
	case GL_RGBA4:
		lightmap_bytes = 2;
		break;
	case GL_LUMINANCE:
	case GL_INTENSITY:
	case GL_ALPHA:
		lightmap_bytes = 1;
		break;
	}

	for (j=1 ; j<MAX_MODELS ; j++)
	{
		m = cl.model_precache[j];
		if (!m)
			break;
		if (m->name[0] == '*')
			continue;
		r_pcurrentvertbase = m->vertexes;
		currentmodel = m;
		for (i=0 ; i<m->numsurfaces ; i++)
		{
			GL_CreateSurfaceLightmap (m->surfaces + i);
			if ( m->surfaces[i].flags & SURF_DRAWTURB )
				continue;
#ifndef QUAKE2
			if ( m->surfaces[i].flags & SURF_DRAWSKY )
				continue;
#endif
			BuildSurfaceDisplayList (m->surfaces + i);
		}
	}

 	if (!gl_texsort.value)
 		GL_SelectTexture(TEXTURE1_SGIS);

	//
	// upload all lightmaps that were filled
	//
	for (i=0 ; i<MAX_LIGHTMAPS ; i++)
	{
		if (!allocated[i][0])
			break;		// no more used
		lightmap_modified[i] = false;
		lightmap_rectchange[i].l = BLOCK_WIDTH;
		lightmap_rectchange[i].t = BLOCK_HEIGHT;
		lightmap_rectchange[i].w = 0;
		lightmap_rectchange[i].h = 0;
		GL_Bind(lightmap_textures + i);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexImage2D (GL_TEXTURE_2D, 0, lightmap_bytes
		, BLOCK_WIDTH, BLOCK_HEIGHT, 0, 
		gl_lightmap_format, GL_UNSIGNED_BYTE, lightmaps+i*BLOCK_WIDTH*BLOCK_HEIGHT*lightmap_bytes);
	}

 	if (!gl_texsort.value)
 		GL_SelectTexture(TEXTURE0_SGIS);

}

