//rww - danger zone stuff, shared between client and server

#ifdef SERVERONLY
	#include "../server/qwsvdef.h"
#else
	#include "quakedef.h"
#endif

#include "qr_dzone.h"

extern float g_worldMins[3];
extern float g_worldMaxs[3];

qrDZone_t g_qrDZones[TOTAL_DZONES];
float g_dZoneSize[2];

//see if a position is in an active danger zones
qboolean QR_InDangerZone(float *pos)
{
	int i, j;
	for (i = 0; i < TOTAL_DZONES; i++)
	{
		qrDZone_t *dz = &g_qrDZones[i];
		if (!dz->active)
		{
			continue;
		}

		for (j = 0; j < 3; j++)
		{
			if (pos[j] < dz->mins[j] || pos[j] > dz->maxs[j])
			{
				break;
			}
		}
		if (j == 3)
		{
			return true;
		}
	}

	return false;
}

#ifdef SERVERONLY //server-only
//sends updates to the client
void QR_NetUpdateZones(client_t *cl)
{
	int i;
	ClientReliableWrite_Begin(cl, svc_dzones, TOTAL_DZONES+1);
	for (i = 0; i < TOTAL_DZONES; i++)
	{
		qrDZone_t *dz = &g_qrDZones[i];
		if (dz->active)
		{
			ClientReliableWrite_Byte(cl, 2);
		}
		else if (dz->pending)
		{
			ClientReliableWrite_Byte(cl, 1);
		}
		else
		{
			ClientReliableWrite_Byte(cl, 0);
		}
	}
}

void QR_NetUpdateZonesAllClients(void)
{
	int i;
	client_t *client;
	for (i=0, client = svs.clients ; i<MAX_CLIENTS ; i++, client++)
	{
		if ( client->state == cs_spawned )
		{
			QR_NetUpdateZones(client);
		}
	}
}

//deactivate all zones
void QR_DeactivateZones(void)
{
	int i;
	for (i = 0; i < TOTAL_DZONES; i++)
	{
		qrDZone_t *dz = &g_qrDZones[i];
		dz->active = false;
		dz->pending = false;
	}

	QR_NetUpdateZonesAllClients();
}

//get the friendly zone name
char *QR_NameZone(qrDZone_t *dz)
{
	static int i = 0;
	static char str[6][128];
	char *p = str[i];

	sprintf(p, "%c%i", 'a'+(byte)dz->row, dz->col+1);

	i = (i+1)%6;

	return p;
}

//activates pending zones, and sets more pending
void QR_IncrementZones(void)
{
	int i;
	qrDZone_t *inactive[TOTAL_DZONES];
	qrDZone_t *newPend = NULL;
	qrDZone_t *newAct = NULL;
	int numInactive = 0;
	for (i = 0; i < TOTAL_DZONES; i++)
	{
		qrDZone_t *dz = &g_qrDZones[i];
		if (dz->pending)
		{
			dz->active = true;
			dz->pending = false;
			newAct = dz;
		}
		if (!dz->active)
		{
			inactive[numInactive++] = dz;
		}
	}
	
	if (numInactive)
	{
		int r = rand()%numInactive;
		inactive[r]->pending = true;
		newPend = inactive[r];
	}

	QR_NetUpdateZonesAllClients();

	if (newPend && newAct)
	{
		SV_BroadcastPrintf(PRINT_RADAR, "Zone %s activated, %s pending.\n", QR_NameZone(newAct), QR_NameZone(newPend));
	}
	else if (newPend)
	{
		SV_BroadcastPrintf(PRINT_RADAR, "Zone %s now pending.\n", QR_NameZone(newPend));
	}
	else if (newAct)
	{
		SV_BroadcastPrintf(PRINT_RADAR, "Zone %s now active.\n", QR_NameZone(newAct));
	}
}
#else //client-only
void QR_ParseDZones(void)
{
	int i;
	for (i = 0; i < TOTAL_DZONES; i++)
	{
		qrDZone_t *dz = &g_qrDZones[i];
		int a = MSG_ReadByte();
		if (a == 2)
		{
			dz->active = true;
			dz->pending = false;
		}
		else if (a == 1)
		{
			dz->active = false;
			dz->pending = true;
		}
		else
		{
			dz->active = false;
			dz->pending = false;
		}
	}
}
#endif

//create the danger zones using the map's extents
void QR_CreateDangerZones(void)
{
	int i, j;
	float dZoneBase[3];
	g_dZoneSize[0] = (g_worldMaxs[0]-g_worldMins[0])/DZONES_PER_ROW;
	g_dZoneSize[1] = (g_worldMaxs[1]-g_worldMins[1])/DZONES_PER_ROW;

	dZoneBase[0] = g_worldMins[0];
	dZoneBase[1] = g_worldMins[1];
	dZoneBase[2] = 0.0f;

	for (i = 0; i < DZONES_PER_ROW; i++)
	{
		for (j = 0; j < DZONES_PER_ROW; j++)
		{
			qrDZone_t *dz = &g_qrDZones[i*DZONES_PER_ROW + j];
			memset(dz, 0, sizeof(qrDZone_t));

			dz->row = i;
			dz->col = j;

			dz->mins[0] = dZoneBase[0];
			dz->mins[1] = dZoneBase[1];
			dz->mins[2] = dZoneBase[2]-8192.0f;
			dz->maxs[0] = dZoneBase[0]+g_dZoneSize[0];
			dz->maxs[1] = dZoneBase[1]+g_dZoneSize[1];
			dz->maxs[2] = dZoneBase[2]+8192.0f;

			dZoneBase[0] += g_dZoneSize[0];
		}
		dZoneBase[0] = g_worldMins[0];
		dZoneBase[1] += g_dZoneSize[1];
	}

}
