
/* This is an example of how to use libpng to read and write PNG files.
 * The file libpng.txt is much more verbose then this.  If you have not
 * read it, do so first.  This was designed to be a starting point of an
 * implementation.  This is not officially part of libpng, is hereby placed
 * in the public domain, and therefore does not require a copyright notice.
 *
 * This file does not currently compile, because it is missing certain
 * parts, like allocating memory to hold an image.  You will have to
 * supply these parts to get it to compile.  For an example of a minimal
 * working PNG reader/writer, see pngtest.c, included in this distribution;
 * see also the programs in the contrib directory.
 */

#include "png.h"
//rww begin
#include <stdlib.h>
//rww end

 /* The png_jmpbuf() macro, used in error handling, became available in
  * libpng version 1.0.6.  If you want to be able to run your code with older
  * versions of libpng, you must define the macro yourself (but only if it
  * is not already defined by libpng!).
  */

#ifndef png_jmpbuf
#  define png_jmpbuf(png_ptr) ((png_ptr)->jmpbuf)
#endif

/* Check to see if a file is a PNG file using png_sig_cmp().  png_sig_cmp()
 * returns zero if the image is a PNG and nonzero if it isn't a PNG.
 *
 * The function check_if_png() shown here, but not used, returns nonzero (true)
 * if the file can be opened and is a PNG, 0 (false) otherwise.
 *
 * If this call is successful, and you are going to keep the file open,
 * you should call png_set_sig_bytes(png_ptr, PNG_BYTES_TO_CHECK); once
 * you have created the png_ptr, so that libpng knows your application
 * has read that many bytes from the start of the file.  Make sure you
 * don't call png_set_sig_bytes() with more than 8 bytes read or give it
 * an incorrect number of bytes read, or you will either have read too
 * many bytes (your fault), or you are telling libpng to read the wrong
 * number of magic bytes (also your fault).
 *
 * Many applications already read the first 2 or 4 bytes from the start
 * of the image to determine the file type, so it would be easiest just
 * to pass the bytes to png_sig_cmp() or even skip that if you know
 * you have a PNG file, and call png_set_sig_bytes().
 */
#define PNG_BYTES_TO_CHECK 4
int check_if_png(char *file_name, FILE **fp)
{
   char buf[PNG_BYTES_TO_CHECK];

   /* Open the prospective PNG file. */
   if ((*fp = zl_fopen(file_name, "rb")) == NULL)
      return 0;

   /* Read in some of the signature bytes */
   if (zl_fread(buf, 1, PNG_BYTES_TO_CHECK, *fp) != PNG_BYTES_TO_CHECK)
      return 0;

   /* Compare the first PNG_BYTES_TO_CHECK bytes of the signature.
      Return nonzero (true) if they match */

   return(!png_sig_cmp(buf, (png_size_t)0, PNG_BYTES_TO_CHECK));
}

/* Read a PNG file.  You may want to return an error code if the read
 * fails (depending upon the failure).  There are two "prototypes" given
 * here - one where we are given the filename, and we need to open the
 * file, and the other where we are given an open file (possibly with
 * some or all of the magic bytes read - see comments above).
 */

png_voidp user_error_ptr;
png_voidp user_error_fn;
png_voidp user_warning_fn;

png_structp g_png_ptr;
png_infop g_info_ptr;
FILE *fp = NULL;

//rww begin
static unsigned char *activePNGBuffer = NULL;
static int activePNGBufferSize = 0;
static int activePNGBufferOffset = 0;

void PNGAPI PNGActiveBufferReadData(png_structp png_ptr, png_bytep data, png_size_t length)
{
   if ((activePNGBufferOffset+(int)length) > activePNGBufferSize)
   {
	   png_error(png_ptr, "Read Error");
   }

   memcpy(data, activePNGBuffer+activePNGBufferOffset, length);
   activePNGBufferOffset += (int)length;
}

int read_png_buffer(unsigned char *buffer, int size)
{
	unsigned int sig_read = 0;
	if (activePNGBuffer)
	{
		return 0;
	}

   g_png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
      user_error_ptr, user_error_fn, user_warning_fn);

   if (g_png_ptr == NULL)
   {
      return 0;
   }

   /* Allocate/initialize the memory for image information.  REQUIRED. */
   g_info_ptr = png_create_info_struct(g_png_ptr);
   if (g_info_ptr == NULL)
   {
      png_destroy_read_struct(&g_png_ptr, png_infopp_NULL, png_infopp_NULL);
      return 0;
   }

   /* Set error handling if you are using the setjmp/longjmp method (this is
    * the normal method of doing things with libpng).  REQUIRED unless you
    * set up your own error handlers in the png_create_read_struct() earlier.
    */

   if (setjmp(png_jmpbuf(g_png_ptr)))
   {
      /* Free all of the memory associated with the g_png_ptr and info_ptr */
      png_destroy_read_struct(&g_png_ptr, &g_info_ptr, png_infopp_NULL);
      /* If we get here, we had a problem reading the file */
      return 0;
   }

   /* If you are using replacement read functions, instead of calling
    * png_init_io() here you would call:
    */
   activePNGBuffer = buffer;
   activePNGBufferSize = size;
   activePNGBufferOffset = 0;
   png_set_read_fn(g_png_ptr, /*(void *)user_io_ptr*/0, PNGActiveBufferReadData);
   /* where user_io_ptr is a structure you want available to the callbacks */


   /* If we have already read some of the signature */
   png_set_sig_bytes(g_png_ptr, sig_read);

   /*
    * If you have enough memory to read in the entire image at once,
    * and you need to specify only transforms that can be controlled
    * with one of the PNG_TRANSFORM_* bits (this presently excludes
    * dithering, filling, setting background, and doing gamma
    * adjustment), then you can read the entire image (including
    * pixels) into the info structure with this call:
    */
   png_read_png(g_png_ptr, g_info_ptr, /*png_transforms*/0, png_voidp_NULL);

   /* At this point you have read the entire image */

   /* that's it */
   return 1;
}
//rww end

int read_png(const char *file_name)  /* We need to open the file */
{
   unsigned int sig_read = 0;
//   png_uint_32 width, height;
  // int bit_depth, color_type, interlace_type;

   if (fp)
   {
	   return 0;
   }

   if ((fp = zl_fopen(file_name, "rb")) == NULL)
      return 0;

   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also supply the
    * the compiler header file version, so that we know if the application
    * was compiled with a compatible version of the library.  REQUIRED
    */
   g_png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
      user_error_ptr, user_error_fn, user_warning_fn);

   if (g_png_ptr == NULL)
   {
      zl_fclose(fp);
	  fp = NULL;
      return 0;
   }

   /* Allocate/initialize the memory for image information.  REQUIRED. */
   g_info_ptr = png_create_info_struct(g_png_ptr);
   if (g_info_ptr == NULL)
   {
      zl_fclose(fp);
	  fp = NULL;
      png_destroy_read_struct(&g_png_ptr, png_infopp_NULL, png_infopp_NULL);
      return 0;
   }

   /* Set error handling if you are using the setjmp/longjmp method (this is
    * the normal method of doing things with libpng).  REQUIRED unless you
    * set up your own error handlers in the png_create_read_struct() earlier.
    */

   if (setjmp(png_jmpbuf(g_png_ptr)))
   {
      /* Free all of the memory associated with the g_png_ptr and info_ptr */
      png_destroy_read_struct(&g_png_ptr, &g_info_ptr, png_infopp_NULL);
      zl_fclose(fp);
	  fp = NULL;
      /* If we get here, we had a problem reading the file */
      return 0;
   }

   /* If you are using replacement read functions, instead of calling
    * png_init_io() here you would call:
    */
   png_set_read_fn(g_png_ptr, /*(void *)user_io_ptr*/fp, /*user_read_fn*/NULL);
   /* where user_io_ptr is a structure you want available to the callbacks */


   /* If we have already read some of the signature */
   png_set_sig_bytes(g_png_ptr, sig_read);

   /*
    * If you have enough memory to read in the entire image at once,
    * and you need to specify only transforms that can be controlled
    * with one of the PNG_TRANSFORM_* bits (this presently excludes
    * dithering, filling, setting background, and doing gamma
    * adjustment), then you can read the entire image (including
    * pixels) into the info structure with this call:
    */
   png_read_png(g_png_ptr, g_info_ptr, /*png_transforms*/0, png_voidp_NULL);

   /* At this point you have read the entire image */

   /* that's it */
   return 1;
}

void close_png(void)
{
   if (!fp && !activePNGBuffer) //rww - support for buffer ver
   {
	   return;
   }

   /* clean up after the read, and free any memory allocated - REQUIRED */
   png_destroy_read_struct(&g_png_ptr, &g_info_ptr, png_infopp_NULL);

   /* close the file */
   if (fp) //rww - support for buffer ver
   {
	   zl_fclose(fp);
       fp = NULL;
   }

   if (activePNGBuffer) //rww - support for buffer ver
   {
	   activePNGBuffer = NULL;
	   activePNGBufferSize = 0;
	   activePNGBufferOffset = 0;
   }
}

/* progressively read a file */

info_callback(png_structp png_ptr, png_infop info)
{
/* do any setup here, including setting any of the transformations
 * mentioned in the Reading PNG files section.  For now, you _must_
 * call either png_start_read_image() or png_read_update_info()
 * after all the transformations are set (even if you don't set
 * any).  You may start getting rows before png_process_data()
 * returns, so this is your last chance to prepare for that.
 */
}

row_callback(png_structp png_ptr, png_bytep new_row,
   png_uint_32 row_num, int pass)
{
/*
 * This function is called for every row in the image.  If the
 * image is interlaced, and you turned on the interlace handler,
 * this function will be called for every row in every pass.
 *
 * In this function you will receive a pointer to new row data from
 * libpng called new_row that is to replace a corresponding row (of
 * the same data format) in a buffer allocated by your application.
 * 
 * The new row data pointer new_row may be NULL, indicating there is
 * no new data to be replaced (in cases of interlace loading).
 * 
 * If new_row is not NULL then you need to call
 * png_progressive_combine_row() to replace the corresponding row as
 * shown below:
 */
	unsigned int height = png_ptr->height;
    png_bytep old_row = NULL;

   /* Check if row_num is in bounds. */
   if((row_num >= 0) && (row_num < height))
   {
     /* Get pointer to corresponding row in our
      * PNG read buffer.
      */
//     old_row = ((png_bytep *)our_data)[row_num];
//A?WR?!#!
     /* If both rows are allocated then copy the new row
      * data to the corresponding row data.
      */
     if((old_row != NULL) && (new_row != NULL))
     png_progressive_combine_row(png_ptr, old_row, new_row);
   }
/*
 * The rows and passes are called in order, so you don't really
 * need the row_num and pass, but I'm supplying them because it
 * may make your life easier.
 *
 * For the non-NULL rows of interlaced images, you must call
 * png_progressive_combine_row() passing in the new row and the
 * old row, as demonstrated above.  You can call this function for
 * NULL rows (it will just return) and for non-interlaced images
 * (it just does the png_memcpy for you) if it will make the code
 * easier.  Thus, you can just do this for all cases:
 */

   png_progressive_combine_row(png_ptr, old_row, new_row);

/* where old_row is what was displayed for previous rows.  Note
 * that the first pass (pass == 0 really) will completely cover
 * the old row, so the rows do not have to be initialized.  After
 * the first pass (and only for interlaced images), you will have
 * to pass the current row as new_row, and the function will combine
 * the old row and the new row.
 */
}

end_callback(png_structp png_ptr, png_infop info)
{
/* this function is called when the whole image has been read,
 * including any chunks after the image (up to and including
 * the IEND).  You will usually have the same info chunk as you
 * had in the header, although some data may have been added
 * to the comments and time fields.
 *
 * Most people won't do much here, perhaps setting a flag that
 * marks the image as finished.
 */
}



int
initialize_png_reader(png_structp *png_ptr, png_infop *info_ptr)
{
   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also check that
    * the library version is compatible in case we are using dynamically
    * linked libraries.
    */
   *png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
       user_error_ptr, user_error_fn, user_warning_fn);

   if (*png_ptr == NULL)
   {
      *info_ptr = NULL;
      return 0;
   }
#pragma warning (disable : 4047)
#pragma warning (disable : 4024)
   *info_ptr = png_create_info_struct(png_ptr);
#pragma warning (default : 4047)
#pragma warning (default : 4024)

   if (*info_ptr == NULL)
   {
      png_destroy_read_struct(png_ptr, info_ptr, png_infopp_NULL);
      return 0;
   }

   if (setjmp(png_jmpbuf((*png_ptr))))
   {
      png_destroy_read_struct(png_ptr, info_ptr, png_infopp_NULL);
      return 0;
   }

   /* This one's new.  You will need to provide all three
    * function callbacks, even if you aren't using them all.
    * If you aren't using all functions, you can specify NULL
    * parameters.  Even when all three functions are NULL,
    * you need to call png_set_progressive_read_fn().
    * These functions shouldn't be dependent on global or
    * static variables if you are decoding several images
    * simultaneously.  You should store stream specific data
    * in a separate struct, given as the second parameter,
    * and retrieve the pointer from inside the callbacks using
    * the function png_get_progressive_ptr(png_ptr).
    */
   png_set_progressive_read_fn(*png_ptr, NULL,//(void *)stream_data,
      info_callback, row_callback, end_callback);

   return 1;
}

int
process_data(png_structp *png_ptr, png_infop *info_ptr,
   png_bytep buffer, png_uint_32 length)
{
   if (setjmp(png_jmpbuf((*png_ptr))))
   {
      /* Free the png_ptr and info_ptr memory on error */
      png_destroy_read_struct(png_ptr, info_ptr, png_infopp_NULL);
      return 0;
   }

   /* This one's new also.  Simply give it chunks of data as
    * they arrive from the data stream (in order, of course).
    * On Segmented machines, don't give it any more than 64K.
    * The library seems to run fine with sizes of 4K, although
    * you can give it much less if necessary (I assume you can
    * give it chunks of 1 byte, but I haven't tried with less
    * than 256 bytes yet).  When this function returns, you may
    * want to display any rows that were generated in the row
    * callback, if you aren't already displaying them there.
    */
   png_process_data(*png_ptr, *info_ptr, buffer, length);
   return 1;
}

//rww begin
static unsigned char *activePNGWriteBuffer = NULL;
static int activePNGWriteBufferSize = 0;
static int activePNGWriteBufferOffset = 0;

void PNGAPI PNGActiveBufferWriteData(png_structp png_ptr, png_bytep data, png_size_t length) {
   if ((activePNGWriteBufferOffset+(int)length) > activePNGWriteBufferSize)
   {
	   png_error(png_ptr, "Write Error");
   }

   memcpy(activePNGWriteBuffer+activePNGWriteBufferOffset, data, length);
   activePNGWriteBufferOffset += (int)length;
}

int png_write(unsigned char *buffer, int w, int h, int bpp, void *out, int outLen, int hasAlpha, int writeAlpha)
{
	png_structp png_ptr;
	png_infop info_ptr;
	int i, j;
	int bytesPerPixel;
	int writeBytesPerPixel;
	FILE *fp = 0;
	int colorType;
	unsigned char *line;

	if (hasAlpha)
	{
		bytesPerPixel = 4;
	}
	else
	{
		bytesPerPixel = 3;
	}

	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
		(png_voidp)NULL, NULL, NULL);
	if (!png_ptr)
	{
		return -1;
	}

	if (outLen != -1)
	{ //not a file
		if (activePNGWriteBuffer)
		{
			return -1;
		}
		activePNGWriteBuffer = out;
		activePNGWriteBufferSize = outLen;
		activePNGWriteBufferOffset = 0;

		png_set_write_fn(png_ptr, png_voidp_NULL, PNGActiveBufferWriteData,
			png_flush_ptr_NULL);
	}
	else
	{
		fp = zl_fopen((char *)out, "wb+");
		if (!fp)
		{
			return -1;
		}
	}

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) 
	{
		if (activePNGWriteBuffer)
		{
			activePNGWriteBuffer = 0;
		}
		return -1;
	}

	if (setjmp(png_jmpbuf(png_ptr)))
	{
		if (activePNGWriteBuffer)
		{
			activePNGWriteBuffer = 0;
		}
		return -1;
	}

	png_init_io(png_ptr, fp);

	if (writeAlpha)
	{
		colorType = PNG_COLOR_TYPE_RGB_ALPHA;
		writeBytesPerPixel = 4;
	}
	else
	{
		colorType = PNG_COLOR_TYPE_RGB;
		writeBytesPerPixel = 3;
	}

	png_set_compression_level(png_ptr, Z_BEST_SPEED);
	png_set_IHDR(png_ptr, info_ptr,
		w,
		h,
		bpp,
		colorType,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT,
		PNG_FILTER_TYPE_DEFAULT);

	png_write_info(png_ptr, info_ptr);

	line = (unsigned char *)malloc(w*writeBytesPerPixel);

	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++) {
			line[j*writeBytesPerPixel] = buffer[0];
			line[(j*writeBytesPerPixel)+1] = buffer[1];
			line[(j*writeBytesPerPixel)+2] = buffer[2];
			if (writeAlpha && hasAlpha)
			{
				line[(j*writeBytesPerPixel)+3] = buffer[3];
			}
			buffer += bytesPerPixel;
		}
		png_write_row(png_ptr, line);
	}
	png_write_end(png_ptr, info_ptr);

	png_destroy_write_struct(&png_ptr, &info_ptr);

	free(line);

	if (fp)
	{
		zl_fclose(fp);
		fp = NULL;
	}

	if (activePNGWriteBuffer)
	{
		activePNGWriteBuffer = NULL;
		return activePNGWriteBufferOffset;
	}

	return 1;
}
//rww end
